import logging
from dataclasses import asdict, dataclass, field
from typing import Any, Self

from rics.paths import AnyPath, any_path_to_path

from .._load_toml import load_toml_file

LOGGER = logging.getLogger(__package__).getChild("Translator").getChild("config")


@dataclass(frozen=True)
class EnvConf:
    """Control environment-variable interpolation.

    All keys are forwarded to :func:`.load_toml_file`.
    """

    allow_interpolation: bool = True
    allow_blank: bool = False
    allow_nested: bool = False

    @classmethod
    def from_dict(cls, config: dict[str, bool]) -> Self:
        """Construct environment interpolation configuration object from a dict.

        Args:
            config: Dict representation of a ``EnvConf`` to consume.

        Returns:
            A new ``EnvConf`` instance.
        """
        return cls(**config)


@dataclass(frozen=True)
class EquivalenceConf:
    """Determines how equivalence between configuration files is determined."""

    python_version: str = "{v.major}.{v.minor}.{v.micro}"
    """Used to format the ``'python_version'`` version using ``v=sys.version_info``."""
    extra_packages: list[str] = field(default_factory=list)
    """Additional package whose versions are used to determine :class:`ConfigMetadata` equivalence."""

    @classmethod
    def from_dict(cls, config: dict[str, Any]) -> Self:
        """Construct equivalence configuration object from a dict.

        Args:
            config: Dict representation of an ``EquivalenceConf`` to consume.

        Returns:
            A new ``EquivalenceConf`` instance.
        """
        extra_packages = config.pop("extra_packages", [])
        return cls(**config, extra_packages=[*extra_packages])


@dataclass(frozen=True)
class Metaconf:
    """Python representation of ``metaconf.toml`` file.

    The ``metaconf.toml`` file always lives next to the main configuration file. It determines how other configuration
    files are read and validated.

    .. code-block:: toml

        [env]
        allow_interpolation = true
        allow_nested = true
        allow_blank = false

        [equivalence]
        # python_version = "{v.major}.{v.minor}"  # Uncomment to allow any patch version
        extra_packages = ["bci-id-translation"]

    The sample above comes from the https://github.com/rsundqvist/id-translation-project/tree/master/demo/bci-id-translation demo.

    """

    env: EnvConf = field(default_factory=EnvConf)
    """Controls how and if variable substitution (on the form ``${VAR}``) is performed. See :class:`EnvConf`. """

    equivalence: EquivalenceConf = field(default_factory=EquivalenceConf)
    """Controls how equivalence checks are performed. See :class:`EquivalenceConf`."""

    @classmethod
    def from_dict(cls, config: dict[str, Any]) -> Self:
        """Construct meta configuration object from a dict.

        Args:
            config: Dict representation of a ``Metaconf`` to consume.

        Returns:
            A new ``Metaconf`` instance.
        """
        env = config.pop("env", {})
        equivalence = config.pop("equivalence", {})

        if config:
            raise TypeError(f"Unknown keys: {sorted(config)}")

        return cls(
            env=EnvConf.from_dict(env),
            equivalence=EquivalenceConf.from_dict(equivalence),
        )

    @classmethod
    def from_path_or_default(cls, path: AnyPath) -> Self:
        """Read TOML configuration or return default configuration."""
        path = any_path_to_path(path)
        config = load_toml_file(path) if path.exists() else {}
        return cls.from_dict(config)

    def as_dict(self) -> dict[str, Any]:
        """Return a dict representation of `self`."""
        return asdict(self)



"""TOML configuration metadata types."""

from ._base_metadata import BaseMetadata
from ._config_metadata import ConfigMetadata
from ._metaconf import EnvConf, EquivalenceConf, Metaconf

__all__ = [
    "BaseMetadata",
    "ConfigMetadata",
    "EnvConf",
    "EquivalenceConf",
    "Metaconf",
]



"""Logging utilities; see :doc:`/documentation/translation-logging` for help."""

import logging as _l
import typing as _t
from random import Random as _Random
from time import perf_counter as _perf_counter

from rics.env import read as _env

ENABLE_VERBOSE_LOGGING: bool = _env.read_bool("ID_TRANSLATION_VERBOSE")
"""Set to enable additional ``DEBUG``-level messages.

The default (``False``) is controlled by the :envvar:`ID_TRANSLATION_VERBOSE` variable. Use
:func:`enable_verbose_debug_messages` to enable temporarily.
"""

LOGGER = _l.getLogger("id_translation")
"""Namespace root logger.

Level is set to ``logging.WARNING=30`` by default (or ``DEBUG`` if ``ENABLE_VERBOSE_LOGGING`` is set).
See :doc:`/documentation/translation-logging` for details.
"""
if LOGGER.level == _l.NOTSET:
    LOGGER.setLevel(_l.DEBUG if ENABLE_VERBOSE_LOGGING else _l.WARNING)


def enable_verbose_debug_messages(
    level: _t.Literal["verbose", "debug", "info", "warning"] = "verbose",
    *,
    use_custom_handler: bool | _t.Literal["auto"] = "auto",
    style: _t.Literal["minimal", "basic", "pretty", "rainbow"] = "pretty",
) -> _t.ContextManager[None]:
    """Enable verbose logging. May be used as a context.

    .. hint::

       Click `here <../_static/logging-style-rainbow.html>`__ for verbose sample output using ``style="rainbow"``.

    **Styles**

    * **minimal**: Nothing but the logged message itself.
    * **basic**: Adds logger name and level.
    * **pretty**: Adds basic color and task ID (e.g. ``🍏 0x2b92``). Mark stage (``enter=🚀, exit=✅``).
    * **rainbow**: Indent based on stage. Full-color syntax highlighting for strings and keywords.

    Args:
        level: Log level. If `'verbose'` (default), set :attr:`ENABLE_VERBOSE_LOGGING` to ``True``.
        use_custom_handler: Set to ``False`` to use existing handlers. If `'auto'` (default), use existing handlers if
            one is found (see :py:meth:`logging.Logger.hasHandlers`). If ``True``, propagation is disabled for the
            :data:`namespace root logger <LOGGER>`.
        style: Formatting style to use. Ignored when `use_custom_handler` evaluates to ``False``.

    Examples:
        Basic usage.

        >>> from id_translation.mapping import Mapper
        >>> with enable_verbose_debug_messages():
        ...     # Context manager; changes are temporary.
        ...     Mapper().apply("ab", candidates="abc")

        Forcing custom handlers. These add formatting (e.g. color) to namespace logger messages.

        >>> enable_verbose_debug_messages(use_custom_handler=True, style="rainbow")
        >>> Mapper().apply("ab", candidates="abc")

        The changes aren't automatically undone if a regular function call is used.

    Notes:
        Custom handlers emit to standard out.
    """
    global ENABLE_VERBOSE_LOGGING  # noqa: PLW0603

    if level.lower() == "verbose":
        level = "debug"
        verbose = True
    else:
        verbose = False

    verbose_before = ENABLE_VERBOSE_LOGGING
    level_before = LOGGER.level
    propagate_before = LOGGER.propagate

    LOGGER.setLevel(_l._nameToLevel[level.upper()])
    ENABLE_VERBOSE_LOGGING = verbose

    if use_custom_handler == "auto" and not LOGGER.hasHandlers():
        use_custom_handler = True

    handler: _l.Handler | None = None
    if use_custom_handler is True:
        from sys import stdout  # noqa: PLC0415

        if style == "minimal":
            formatter = _l.Formatter("%(message)s")
        elif style == "basic":
            formatter = _l.Formatter("[%(name)s:%(levelname)s] %(message)s")
        else:
            from ._utils.debug_logging_formatter import DebugLoggingFormatter  # noqa: PLC0415

            formatter = DebugLoggingFormatter(less=style == "pretty", indent_style=".." if verbose else "")

        handler = _l.StreamHandler(stdout)
        handler.setFormatter(formatter)
        LOGGER.addHandler(handler)
        LOGGER.propagate = False  # Prevent duplicate output.

    def undo() -> None:
        """Restore original state."""
        global ENABLE_VERBOSE_LOGGING  # noqa: PLW0603

        LOGGER.setLevel(level_before)
        ENABLE_VERBOSE_LOGGING = verbose_before

        if handler:
            LOGGER.propagate = propagate_before
            LOGGER.removeHandler(handler)

    class Undo(_t.ContextManager[None]):
        def __exit__(self, *_: _t.Any) -> None:
            undo()

    return Undo()


def generate_task_id(seed: float | None = None) -> int:
    """Generate a new task ID."""
    if seed is None:
        seed = _perf_counter()
    random = _Random(seed)  # noqa: S311
    return random.randint(0, 65535)  # 65535 = 0xFFFF


def get_event_key(method: _t.Any, stage: str) -> str:
    """Construct `event_key` value."""
    cls = type(method.__self__).__name__
    return f"{cls}.{method.__name__}:{stage}"



"""Default :class:`.DataStructureIO` implementations."""

from ._dict import DictIO
from ._sequence import SequenceIO
from ._set import SetIO
from ._single_value import SingleValueIO

__all__ = [
    "DictIO",
    "SequenceIO",
    "SetIO",
    "SingleValueIO",
]



"""Integration for `Polars <https://pola.rs/>`_ types."""

import typing as _t
import warnings as _w
from collections import abc as _abc

import polars as _pl

from id_translation import dio as _dio
from id_translation import types as _tt
from id_translation.dio.exceptions import NotInplaceTranslatableError as _NotInplaceTranslatableError
from id_translation.offline import TranslationMap as _TranslationMap

PolarsT = _t.TypeVar("PolarsT", _pl.DataFrame, _pl.Series)
"""Supported ``polars`` types."""


class PolarsIO(_dio.DataStructureIO[PolarsT, str, _tt.SourceType, _tt.IdType]):
    """Optional IO implementation for ``polars`` types."""

    priority = 1990

    @classmethod
    def handles_type(cls, arg: _t.Any) -> bool:
        if isinstance(arg, _pl.LazyFrame):
            raise NotImplementedError("polars.LazyFrame not supported")
        return isinstance(arg, (_pl.DataFrame, _pl.Series))

    @classmethod
    def names(cls, translatable: PolarsT) -> list[str] | None:
        if isinstance(translatable, _pl.DataFrame):
            return list(translatable.columns)

        return None if translatable.name is None else [translatable.name]

    @staticmethod
    def _obj_to_str(series: _pl.Series) -> _pl.Series:
        if series.dtype == _pl.Object:
            with _w.catch_warnings():
                _w.filterwarnings("ignore", category=_pl.exceptions.PolarsInefficientMapWarning)
                series = series.map_elements(str, return_dtype=_pl.String)
        return series

    @classmethod
    def extract(cls, translatable: PolarsT, names: list[str]) -> dict[str, _abc.Sequence[_tt.IdType]]:
        def extract(series: _pl.Series) -> _abc.Sequence[_tt.IdType]:
            return cls._obj_to_str(series).unique().to_list()

        if isinstance(translatable, _pl.Series):
            if len(names) != 1:
                raise RuntimeError(f"{len(names)=} != 1 is not supported for polars.Series")
            return {names[0]: extract(translatable)}
        else:
            return {n: extract(translatable[n]) for n in names}

    @classmethod
    def insert(
        cls,
        translatable: PolarsT,
        names: list[str],
        tmap: _TranslationMap[str, _tt.SourceType, _tt.IdType],
        copy: bool,
    ) -> PolarsT | None:
        if not copy and isinstance(translatable, _pl.Series):
            raise _NotInplaceTranslatableError(translatable)

        def _translate_series(series: _pl.Series, name: str) -> _pl.Series:
            # Create the mappings before Polars can disappear into Rust, where the MagicDict logic will disappear. For
            # the base case this is fine, but things like Transformer.try_add_missing_key will break.
            magic_dict = tmap[name]
            series = cls._obj_to_str(series)
            mapping = {idx: magic_dict[idx] for idx in series.unique()}
            return series.replace_strict(mapping, return_dtype=_pl.String)

        if isinstance(translatable, _pl.DataFrame):
            translated_columns = {name: _translate_series(translatable[name], name) for name in names}
            if copy:
                return translatable.with_columns(**translated_columns)
            else:
                for name, column in translated_columns.items():
                    translatable.replace_column(translatable.columns.index(name), column)
                return None
        else:
            return _translate_series(translatable, names[0])



"""Types used for translation.

.. hint::

   Use :func:`.register_io` to register custom :class:`.DataStructureIO` implementations.

Rules of thumb
--------------
* The :attr:`IdTypes` are the only `"truly"` translatable types. Collections thereof are also supported.
* Non-inplace (copy-translation, default) always `tries` to return a collection of the same type, with all
  :attr:`IdTypes` converted to :py:class:`str`.
* In-place translation always returns ``None``, or raises :class:`~.dio.exceptions.NotInplaceTranslatableError`.

Overloads err on the overly-permissive side. Static type checkers (only MyPy is tested) may incorrectly allow things
like ``translator.translate("a-string-id")`` for ``int``-only ``Translator`` instances.

Limitations
-----------
Python does not support generics-of-generics, which is the primary domain in which the ``Translator.translate``-method
operates. See https://github.com/python/typing/issues/548 for details on this subject.

* Numpy is supported, but not typed.
* Pandas typing must be enabled using the ``ID_TRANSLATION_PANDAS_IS_TYPED``-flag.

.. code-block:: console

   pip install pandas-stubs
   mypy --always-true=ID_TRANSLATION_PANDAS_IS_TYPED /path/to/file.py

All ``pandas`` types will be ``None`` without stubs, which will break the overloads.
"""

import abc as _abc
import typing as _t
from collections import abc as _cabc
from typing import TYPE_CHECKING
from uuid import UUID as _UUID

ID: _t.Literal["id"] = "id"
"""Name of the ID placeholder in the translation :class:`~.Format`.

Cannot be changed (yet); see https://github.com/rsundqvist/id-translation/issues/151.
"""

IdTypes = int | str | _UUID
"""Type of the value being translated into human-readable labels."""
IdType = _t.TypeVar("IdType", bound=IdTypes)
"""Type variable bound by :attr:`IdTypes`."""

NameType = _t.TypeVar("NameType", bound=_cabc.Hashable)
"""Type used to label collections of IDs, such as the column names in a DataFrame or the keys of a dict."""

if TYPE_CHECKING:
    import pandas


CopyTranslatable: _t.TypeAlias = (
    # Scalar
    IdType
    |
    # Tuple
    tuple[IdType]
    | tuple[IdType, IdType]
    | tuple[IdType, IdType, IdType]
    | tuple[IdType, ...]
)

_InplaceTranslatable: _t.TypeAlias = list[IdType] | list[list[IdType]] | set[IdType]


# From CopyTranslatable
DictToId = dict[NameType, IdType]
DictToList = dict[NameType, list[IdType]]
DictToSet = dict[NameType, set[IdType]]
DictToOneTuple = dict[NameType, tuple[IdType]]
DictToTwoTuple = dict[NameType, tuple[IdType, IdType]]
DictToThreeTuple = dict[NameType, tuple[IdType, IdType, IdType]]
DictToVarTuple = dict[NameType, tuple[IdType, ...]]
DictTranslatable = (
    DictToId[NameType, IdType]
    | DictToList[NameType, IdType]
    | DictToSet[NameType, IdType]
    | DictToOneTuple[NameType, IdType]
    | DictToTwoTuple[NameType, IdType]
    | DictToThreeTuple[NameType, IdType]
    | DictToVarTuple[NameType, IdType]
)

PandasTranslatable = _t.Union["pandas.DataFrame", "pandas.Series", "pandas.Index"]

InplaceTranslatable: _t.TypeAlias = DictTranslatable[NameType, IdType] | _InplaceTranslatable[IdType]

Translatable: _t.TypeAlias = InplaceTranslatable[NameType, IdType] | CopyTranslatable[IdType] | PandasTranslatable
"""Enumeration of translatable types.

Types ``int``, ``str``, and ``UUID`` can be translated, or a collection thereof. Some :mod:`numpy` and :mod:`pandas`
types are also supported. The :class:`.Translator` is quite flexible when it comes to the encapsulating data structure,
and will do its best to return a data structure of the same type (albeit with elements converted to ``str``).

.. note::

   Dict values are always copied for translation.

   Setting ``copy=False`` merely controls whether the original dict is modified.
"""


SourceType = _t.TypeVar("SourceType", bound=_cabc.Hashable)
"""Type used to describe sources. Typically a string for things like files and database tables."""

NameToSource = dict[NameType, SourceType]
"""A mapping from name to source."""

NamesPredicate = _t.Callable[[NameType], bool]
"""A predicate type on names."""
NameTypes: _t.TypeAlias = NameType | _cabc.Iterable[NameType]
"""A union of a name type, or an iterable thereof."""
Names: _t.TypeAlias = NameTypes[NameType] | NamesPredicate[NameType]
"""Acceptable name types."""

TranslatableT = _t.TypeVar("TranslatableT", bound=Translatable[_t.Any, _t.Any])  # TODO: Higher-Kinded TypeVars
"""Simplified ``Translatable`` type."""


class HasSources(_abc.ABC, _t.Generic[SourceType]):
    """Indicates that `sources` and `placeholders` are available."""

    @property
    def sources(self) -> list[SourceType]:
        """A list of known Source names, such as ``cities`` or ``languages``."""
        return list(self.placeholders)

    @property
    @_abc.abstractmethod
    def placeholders(self) -> dict[SourceType, list[str]]:
        """Placeholders for all known Source names, such as ``id`` or ``name``.

        These are the (possibly unmapped) placeholders that may be used for translation.

        Returns:
            A dict ``{source: [placeholders..]}``.
        """



import logging
from collections.abc import Iterable
from dataclasses import dataclass
from typing import Generic, Self

from ... import logging as _logging
from .. import Cardinality, DirectionalMapping
from ..exceptions import AmbiguousScoreError
from ..types import CandidateType, ValueType
from ._score_matrix import ScoreMatrix

inf = float("inf")


@dataclass(frozen=True)
class Record(Generic[ValueType, CandidateType]):
    """Data concerning a match."""

    value: ValueType
    """A hashable :class:`value <.ValueType>`."""
    candidate: CandidateType
    """A hashable :class:`candidate <.CandidateType>`."""
    score: float
    """Likeness score computed by some scoring function."""

    def __lt__(self, other: Self) -> bool:
        return self.score < other.score

    def __str__(self) -> str:
        return f"{self.value!r} -> '{self.candidate}'; score={self.score:.3f}"


@dataclass(frozen=True)
class Reject(Generic[ValueType, CandidateType]):
    """Data concerning the rejection of a match."""

    record: Record[ValueType, CandidateType]
    """A :class:`Record` to describe."""
    superseding_value: Record[ValueType, CandidateType] | None = None
    """A :class:`Record` that prevents matching of the current value."""
    superseding_candidate: Record[ValueType, CandidateType] | None = None
    """A :class:`Record` that prevents matching of the current candidate."""

    def explain(self, min_score: float, full: bool = False) -> str:
        """Create a string which explains the rejection.

        Args:
            min_score: Minimum score to accept a match.
            full: If ``True`` show full information about superseding matches.

        Returns:
            An explanatory string.
        """
        if self.record.score == -inf:
            if self.superseding_value and self.superseding_value.score == inf:
                extra = f": {self.superseding_value}" if full else ""
                why = f" (superseded by short-circuit or override{extra})"
            elif self.superseding_candidate and self.superseding_candidate.score == inf:
                extra = f": {self.superseding_candidate}" if full else ""
                why = f" (superseded by short-circuit or override{extra}"
            else:
                why = " (filtered)"
        elif self.record.score < min_score:
            why = f" < {min_score} (below threshold)"
        else:
            ands = []
            if self.superseding_value:
                extra = f": {self.superseding_value}" if full else ""
                ands.append(f"value={self.superseding_value.value!r}{extra}")
            if self.superseding_candidate:
                extra = f": {self.superseding_candidate}" if full else ""
                ands.append(f"candidate={self.superseding_candidate.candidate!r}{extra}")
            why = f" (superseded on {' and '.join(ands)})"

        return f"{self.record}{why}."


class ScoreHelper(Generic[ValueType, CandidateType]):
    """High-level selection operations.

    Args:
        matrix: A :class:`.ScoreMatrix` instance.
        min_score: Minimum score to make a `value -> candidate` match.
        logger: Explicit ``Logger`` instance to use.
        task_id: Used for logging.
    """

    def __init__(
        self,
        matrix: ScoreMatrix[ValueType, CandidateType],
        min_score: float,
        logger: logging.Logger | None = None,
        *,
        task_id: int | None = None,
    ) -> None:
        self._min_score = min_score
        self._matrix = matrix
        self._logger = logging.getLogger(__name__) if logger is None else logger
        self._task_id = task_id

    @property
    def logger(self) -> logging.Logger:
        """Return the ``Logger`` that is used by this instance."""
        return self._logger

    def to_directional_mapping(
        self,
        cardinality: Cardinality | None = None,
    ) -> DirectionalMapping[ValueType, CandidateType]:
        """Create a ``DirectionalMapping`` with a given target ``Cardinality``.

        Args:
            cardinality: Explicit cardinality to set, see :attr:`~.DirectionalMapping.cardinality`. If ``None``, use the
                actual cardinality when selecting all matches with scores :attr:`at or above <above>` the minimum.

        Returns:
            A ``DirectionalMapping``.
        """
        matches: list[Record[ValueType, CandidateType]]
        rejections: list[Reject[ValueType, CandidateType]]
        matches, rejections = self._match(cardinality)
        min_score = self._min_score

        logger = self.logger
        logging_disabled = not (_logging.ENABLE_VERBOSE_LOGGING and logger.isEnabledFor(logging.DEBUG))

        left_to_right: dict[ValueType, list[CandidateType]] = {}
        for record in list(matches):
            left_to_right.setdefault(record.value, []).append(record.candidate)

            if logging_disabled:
                continue

            supersedes: list[Reject[ValueType, CandidateType]] = []
            if rejections:
                supersedes.extend(
                    rr
                    for rr in rejections
                    if record in (rr.superseding_value, rr.superseding_candidate) and rr.record.score >= min_score
                )

            reason = "(short-circuit or override)" if record.score == inf else f">= {min_score}"
            msg = f"Accepted: {record} {reason}."

            if supersedes:
                s = "\n".join("    " + rr.explain(min_score) for rr in supersedes)
                msg += f" This match supersedes {len(supersedes)} other matches:\n{s}"
            logger.debug(msg, extra={"task_id": self._task_id})

        values = set(self._matrix.values)
        if rejections and not logging_disabled:
            unmapped_values = values.difference(left_to_right)
            for value in unmapped_values:
                value_reasons = "\n".join(
                    "    " + reject.explain(min_score, full=True)
                    for reject in rejections
                    if reject.record.value == value
                )
                logger.debug(
                    f"Could not map {value=}. Rejected matches:\n{value_reasons}",
                    extra={"task_id": self._task_id},
                )

        return DirectionalMapping(
            cardinality=cardinality,
            left_to_right={value: tuple(left_to_right[value]) for value in values.intersection(left_to_right)},
            _verify=False,
        )

    def _match(
        self,
        cardinality: Cardinality | None = None,
    ) -> tuple[list[Record[ValueType, CandidateType]], list[Reject[ValueType, CandidateType]]]:
        rejections: list[Reject[ValueType, CandidateType]] | None = None
        records: list[Record[ValueType, CandidateType]] = self.above()

        if _logging.ENABLE_VERBOSE_LOGGING and self.logger.isEnabledFor(logging.DEBUG):
            rejections = []
            records.extend(self.below())

        records.sort(reverse=True)

        if cardinality is Cardinality.OneToOne:
            matches = self._select_one_to_one(records, rejections)
        elif cardinality is Cardinality.OneToMany:
            matches = self._select_one_to_many(records, rejections)
        elif cardinality is Cardinality.ManyToOne:
            matches = self._select_many_to_one(records, rejections)
        else:
            matches = self._select_many_to_many(records, rejections)

        return list(matches), rejections or []

    def above(self) -> list[Record[ValueType, CandidateType]]:
        """Records with scores `above` the threshold."""
        d = self._matrix.to_dict()
        d = {k: v for k, v in d.items() if v >= self._min_score}
        return self._from_dict(d)

    def below(self) -> list[Record[ValueType, CandidateType]]:
        """Records with scores `below` the threshold."""
        d = self._matrix.to_dict()
        d = {k: v for k, v in d.items() if v < self._min_score}
        return self._from_dict(d)

    @classmethod
    def _from_dict(cls, d: dict[tuple[ValueType, CandidateType], float]) -> list[Record[ValueType, CandidateType]]:
        return [Record(value, candidate, score) for (value, candidate), score in d.items()]

    def _raise_if_ambiguous(
        self,
        record: Record,  # type: ignore[type-arg]
        matches: dict,  # type: ignore[type-arg]
        kind: str,
        cardinality: Cardinality,
    ) -> None:
        if record.score == inf:
            # Overrides are allowed to be infinite; the first one will be chosen. It's up to the user to manage them.
            return

        key = record.value if kind == "value" else record.candidate
        if key not in matches:
            return

        old_match = matches[key]
        if old_match.score == inf:
            # Overrides are allowed to be infinite; the first one will be chosen. It's up to the user to manage them.
            return

        if record.score == old_match.score:
            raise AmbiguousScoreError(
                kind=kind,
                key=key,
                match0=record,
                match1=old_match,
                cardinality=cardinality.name,
                scores=self._matrix.to_string(),
            )

    def _select_one_to_one(
        self,
        records: Iterable[Record[ValueType, CandidateType]],
        rejections: list[Reject[ValueType, CandidateType]] | None = None,
    ) -> Iterable[Record[ValueType, CandidateType]]:
        mvs: dict[ValueType, Record[ValueType, CandidateType]] = {}
        mcs: dict[CandidateType, Record[ValueType, CandidateType]] = {}

        for record in records:
            self._raise_if_ambiguous(record, mcs, "candidate", Cardinality.OneToOne)
            self._raise_if_ambiguous(record, mvs, "value", Cardinality.OneToOne)

            if record.score < self._min_score or record.value in mvs or record.candidate in mcs:
                if rejections is not None:
                    rejections.append(
                        Reject(
                            record,
                            superseding_value=mvs.get(record.value),
                            superseding_candidate=mcs.get(record.candidate),
                        )
                    )
                continue
            mvs[record.value] = record
            mcs[record.candidate] = record
            yield record

    def _select_one_to_many(
        self,
        records: Iterable[Record[ValueType, CandidateType]],
        rejections: list[Reject[ValueType, CandidateType]] | None = None,
    ) -> Iterable[Record[ValueType, CandidateType]]:
        mcs: dict[CandidateType, Record[ValueType, CandidateType]] = {}

        for record in records:
            self._raise_if_ambiguous(record, mcs, "candidate", Cardinality.OneToMany)

            if record.score < self._min_score or record.candidate in mcs:
                if rejections is not None:
                    rejections.append(Reject(record, superseding_candidate=mcs.get(record.candidate)))
                continue
            mcs[record.candidate] = record
            yield record

    def _select_many_to_one(
        self,
        records: Iterable[Record[ValueType, CandidateType]],
        rejections: list[Reject[ValueType, CandidateType]] | None = None,
    ) -> Iterable[Record[ValueType, CandidateType]]:
        mvs: dict[ValueType, Record[ValueType, CandidateType]] = {}

        for record in records:
            self._raise_if_ambiguous(record, mvs, "value", cardinality=Cardinality.ManyToOne)

            if record.score < self._min_score or record.value in mvs:
                if rejections is not None:
                    rejections.append(Reject(record, superseding_value=mvs.get(record.value)))
                continue
            mvs[record.value] = record
            yield record

    def _select_many_to_many(
        self,
        records: Iterable[Record[ValueType, CandidateType]],
        rejections: list[Reject[ValueType, CandidateType]] | None = None,
    ) -> Iterable[Record[ValueType, CandidateType]]:
        for record in records:
            if record.score < self._min_score:
                if rejections is not None:
                    rejections.append(Reject(record))
                continue
            yield record



from typing import Any

from rics.collections.dicts import InheritedKeysDict

from id_translation import exceptions
from id_translation.mapping import HeuristicScore, Mapper


def default_mapper_factory(config: dict[str, Any], for_fetcher: bool) -> Mapper[Any, Any, Any] | None:
    """Create a :class:`.Mapper` from config."""
    if "score_function" in config and isinstance(config["score_function"], dict):
        score_function = config.pop("score_function")

        if len(score_function) > 1:  # pragma: no cover
            raise exceptions.ConfigurationError(
                f"At most one score function may be specified, but got: {sorted(score_function)}"
            )

        score_function, score_function_kwargs = next(iter(score_function.items()))
        config["score_function"] = score_function
        config["score_function_kwargs"] = score_function_kwargs

    if "score_function_heuristics" in config:
        if "score_function" not in config:  # pragma: no cover
            section = "fetching" if for_fetcher else "translation"
            raise exceptions.ConfigurationError(
                f"Section [{section}.mapper.score_function_heuristics] requires an explicit score function."
            )

        heuristics = [
            (heuristic_config.pop("function"), heuristic_config)
            for heuristic_config in config.pop("score_function_heuristics")
        ]
        score_function = config["score_function"]

        if isinstance(score_function, HeuristicScore):  # pragma: no cover
            for h, kwargs in heuristics:
                score_function.add_heuristic(h, kwargs)
        else:
            config["score_function"] = HeuristicScore(score_function, heuristics)

    if "filter_functions" in config:
        config["filter_functions"] = [(f.pop("function"), f) for f in config.pop("filter_functions")]

    if "overrides" in config:  # pragma: no cover
        overrides = config.pop("overrides")
        shared, specific = _split_overrides(overrides)

        if specific and not for_fetcher:
            raise exceptions.ConfigurationError(
                "Context-sensitive overrides are not possible (or needed) for "
                f"Name-to-source mapping, but got {overrides=}."
            )

        config["overrides"] = InheritedKeysDict(specific, default=shared) if for_fetcher else shared

    return Mapper(**config)


def _split_overrides(overrides: Any) -> Any:
    specific = {k: v for k, v in overrides.items() if isinstance(v, dict)}
    shared = {k: v for k, v in overrides.items() if k not in specific}
    return shared, specific



import logging
from collections.abc import Callable, Mapping
from pathlib import Path
from typing import Any, Unpack

import pandas as pd
from rics.misc import get_by_full_name, tname
from rics.types import AnyPath

from ..logging import generate_task_id
from ..offline.types import PlaceholderTranslations
from ..translator_typing import AbstractFetcherParams
from ..types import IdType
from ._abstract_fetcher import AbstractFetcher
from .types import FetchInstruction

PandasReadFunction = Callable[[AnyPath], pd.DataFrame]
FormatFn = Callable[[str], str]


class PandasFetcher(AbstractFetcher[str, IdType]):
    """Fetcher implementation using :class:`pandas.DataFrame` as the data format.

    Fetch data from serialized frames. How this is done is determined by the `read_function`. This is typically a Pandas
    function such as :func:`pandas.read_csv` or :func:`pandas.read_parquet`, but any function that accepts a string
    `source` as the first argument and returns a :class:`pandas.DataFrame` can be used.

    .. hint::

       When using **remote file systems**, :attr:`~.AbstractFetcher.sources` are resolved using
       `AbstractFileSystem.glob()`_. If resolution fails, consider overriding the :meth:`find_sources`-method.

    Args:
        read_function: A function ``(str) -> DataFrame``. Derive from `read_path_format` if ``None``. Strings are
            resolved by :func:`~rics.misc.get_by_full_name` (with ``default_module=pandas``).
        read_path_format: A string on the form ``protocol://path/to/sources/{}.<ext>``, or a callable to apply
            to a source before passing them to `read_function`.
        read_function_kwargs: Additional keyword arguments for `read_function`.
        **kwargs: See :class:`.AbstractFetcher`.

    See Also:
        The official `Pandas IO documentation <https://pandas.pydata.org/pandas-docs/stable/user_guide/io.html>`_

    .. _AbstractFileSystem.glob(): https://filesystem-spec.readthedocs.io/en/latest/api.html?highlight=glob#fsspec.spec.AbstractFileSystem.glob
    """

    def __init__(
        self,
        read_function: PandasReadFunction | str | None = None,
        read_path_format: str | FormatFn = "data/{}.csv",
        read_function_kwargs: Mapping[str, Any] | None = None,
        **kwargs: Unpack[AbstractFetcherParams[str, IdType]],
    ) -> None:
        super().__init__(**kwargs)

        self._read = self._derive_read_function(read_function, read_path_format)
        self._format_source: FormatFn = read_path_format if callable(read_path_format) else read_path_format.format
        self._kwargs = read_function_kwargs or {}

        self._source_paths: dict[str, str] = {}

    def read(self, source_path: AnyPath) -> pd.DataFrame:
        """Read a ``DataFrame`` from a source path.

        Args:
            source_path: Path to serialized ``DataFrame``.

        Returns:
            A deserialized ``DataFrame``.
        """
        return self._read(source_path, **self._kwargs).convert_dtypes()

    def format_source(self, source: str) -> str:
        """Get the path for `source`."""
        return self._format_source(source)

    def find_sources(self, task_id: int | None = None) -> dict[str, str]:
        """Resolve sources and their associated paths.

        Args:
            task_id: Used for logging.

        Sources are resolved in three steps:

        1. Create glob pattern by calling :meth:`format_source` with ``source='*'``.
        2. Glob files using `AbstractFileSystem.glob()`_ (requires ``fsspec``) or :meth:`Path.glob() <pathlib.Path.glob>`.
        3. Strip the directory and file suffix from the globbed paths to create source names.

        .. _AbstractFileSystem.glob(): https://filesystem-spec.readthedocs.io/en/latest/api.html?highlight=glob#fsspec.spec.AbstractFileSystem.glob

        Returns:
            A dict ``{source: path}``.
        """
        if task_id is None:
            task_id = generate_task_id()

        pattern = self.format_source("*")

        try:
            sources = self._find_sources_fsspec(pattern)
        except ModuleNotFoundError as e:
            self.logger.debug("Falling back to 'pathlib.Path': %r", e)
            sources = self._find_sources_pathlib(Path(pattern).expanduser().resolve())

        source_paths: dict[str, str] = {source: self.format_source(source) for source in sources}

        extra = {"task_id": task_id, "pattern": pattern, "source_paths": source_paths}
        if not source_paths:
            self.logger.warning(f"Path {pattern=} did not match any files.", extra=extra)
            return {}

        if self.logger.isEnabledFor(logging.DEBUG):
            self.logger.debug(f"Path {pattern=} matched {len(source_paths)} files: {source_paths}", extra=extra)

        return source_paths

    def _find_sources_fsspec(self, pattern: str) -> list[str]:
        from fsspec.core import url_to_fs  # type: ignore  # noqa: PLC0415

        fs, pattern = url_to_fs(pattern, **self._kwargs.get("storage_options", {}))
        return [self._path_to_source(path, pattern) for path in fs.glob(pattern)]

    @classmethod
    def _find_sources_pathlib(cls, pattern: Path) -> list[str]:
        return [cls._path_to_source(str(path), str(pattern)) for path in pattern.parent.glob(pattern.name)]

    @classmethod
    def _path_to_source(cls, path: str, pattern: str) -> str:
        glob_index = pattern.index("*")
        prefix = pattern[:glob_index]
        suffix = pattern[glob_index + 1 :]
        return path.removeprefix(prefix).removesuffix(suffix)

    def _initialize_sources(self, task_id: int) -> dict[str, list[str]]:
        self._source_paths = self.find_sources(task_id)
        return {source: self.read(path).columns.tolist() for source, path in self._source_paths.items()}

    def fetch_translations(self, instr: FetchInstruction[str, IdType]) -> PlaceholderTranslations[str]:
        return PlaceholderTranslations.make(
            instr.source,
            self.read(self._source_paths[instr.source]),
        )

    def __repr__(self) -> str:
        read_path_format = self.format_source("{}")
        return f"{tname(self)}(read_function={tname(self._read)}, {read_path_format=})"

    def _derive_read_function(
        self,
        read_function: PandasReadFunction | str | None,
        read_path_format: str | FormatFn,
    ) -> PandasReadFunction:
        if callable(read_function):
            return read_function

        if read_function is None:
            if not isinstance(read_path_format, str):
                msg = f"Cannot derive `read_function` from {read_path_format=} of type={type(read_function)}."
                raise ValueError(msg)

            func_to_suffixes: dict[PandasReadFunction, list[str]] = {
                pd.read_csv: [".csv"],
                pd.read_pickle: [".pickle", ".pkl"],
                pd.read_feather: [".feather", ".ftr"],
                pd.read_json: [".json"],
                pd.read_parquet: [".parquet", ".parq"],
            }

            for func, suffixes in func_to_suffixes.items():
                for suffix in suffixes:
                    if suffix in read_path_format:  # Using endswith would break paths such as .csv.zip
                        if self.logger.isEnabledFor(logging.INFO):
                            read_function = f"pandas.{func.__name__}"
                            self.logger.debug(
                                f"Derived {read_function=} based on {suffix=} found in {read_path_format=}.",
                                extra={"suffix": suffix, "read_function": read_function},
                            )
                        return func

            suffixes = [suffix for suffixes in func_to_suffixes.values() for suffix in suffixes]
            msg = f"Cannot derive `read_function` from {read_path_format=}; does not contain any known {suffixes=}."
            raise ValueError(msg)

        func = get_by_full_name(read_function, pd)
        if not callable(func):
            msg = f"Bad {read_function=}; type {type(func).__name__} is not callable."
            raise TypeError(msg)
        return func  # type: ignore[no-any-return]



"""Translation of IDs with flexible formatting and name matching.

For an introduction to translation, see :ref:`translation-primer` and :ref:`mapping-primer`.

Environment variables
---------------------
.. envvar:: ID_TRANSLATION_DISABLED

   Global switch. When ``true``, the :meth:`Translator.translate`-method returns immediately.

   .. warning:: When set, IDs are not converted to :py:class:`str`. Type hints may be wrong.

.. envvar:: ID_TRANSLATION_SUPPRESS_OPTIONAL_FETCHER_INIT_ERRORS

   Global switch. Set to ``true`` (not recommended) to allow the :meth:`~id_translation.toml.TranslatorFactory` to
   discard `optional` fetchers that raise when imported or initialized.
   See the `documentation <../documentation/translator-config.html#optional-fetchers>`__ for details.

.. envvar:: ID_TRANSLATION_VERBOSE

   Global switch. Set to ``true`` to overwrite the default :data:`~.logging.ENABLE_VERBOSE_LOGGING` value. Note that
   this variable is only read once (on module import). See the :doc:`documentation </documentation/translation-logging>`
   for details.
"""

from ._translator import Translator

__all__ = [
    "Translator",
    "__version__",  # Make MyPy happy
]

__version__ = "1.0.4"



"""Insertion and extraction of IDs and translations.

See :doc:`/documentation/translation-io` for help.
"""

from ._data_structure_io import DataStructureIO
from ._resolve import (
    _ENTRYPOINT_GROUP,
    get_resolution_order,
    is_registered,
    load_integrations,
    register_io,
    resolve_io,
)

ENTRYPOINT_GROUP: str = _ENTRYPOINT_GROUP  # Public reexport. Makes Sphinx happy.
"""Group used to discover :class:`DataStructureIO` integrations.

See :func:`load_integrations` and :py:func:`importlib.metadata.entry_points` for details.
"""

__all__ = [
    "ENTRYPOINT_GROUP",
    "DataStructureIO",
    "get_resolution_order",
    "is_registered",
    "load_integrations",
    "register_io",
    "resolve_io",
]

load_integrations()



"""Utility for single-purpose translation tasks.

Typing rules
------------

Compared to :meth:`.Translator.translate`, typing is limited. Rules for :meth:`.TranslationHelper.apply`:

* When ``user_params = False``, output= input.
* When ``user_params != False``, output= ``Any`` (new variable) or same (existing variable).
* When ``copy = False``, output= ``None``.

.. warning:: When ``user_params = False``, IDs are not converted to :py:class:`str`. Type hints may be wrong.
.. seealso:: The :envvar:`ID_TRANSLATION_DISABLED` environment variable.

.. _translation-helper-examples:

Examples:
    Implementing a function with a ``translate`` arg using the helper class.

    **Initialization**

    Typically, you'd use something like a :meth:`id_translation.Translator.from_config` callback with suitable
    arguments. Dummy translations are used here.

    >>> from id_translation import Translator
    >>> helper = TranslationHelper[str, str, int](
    ...     Translator,
    ...     "translate",  # for error messages
    ...     names="name",  # fixed translation argument
    ... )
    >>> helper
    TranslationHelper(id_translation.Translator, names='name')

    The `user_params_name='translate'` argument is not printed because it's the default.

    **Function definition**

    Arguments provided when the helper is initialized are fixed. An exception is raised if fixed arguments overlap
    either with `user_kwargs`, or with the defaults provided as keyword-arguments to :meth:`~.TranslationHelper.apply`.

    In the example below, ``names="name"`` is a fixed argument (see the snippet above) and ``fmt="{id}:{name}"`` is a
    default argument.

    >>> def example(
    ...     n: int,
    ...     *,
    ...     translate: UserParams[str, str, int] = True,
    ... ) -> list[str]:
    ...     items: list[str] = helper.apply(
    ...         list(range(n)),
    ...         copy=True,  # required
    ...         user_params=translate,  # forwarded
    ...         fmt="{id}:{name}",  # default - user params can override
    ...     )
    ...     return items

    .. note::

       The `translatable` and `copy` arguments are always required (for :py:func:`@overload <typing.overload>` typing
       reasons).

    Let's take our new function for a spin.

    **Basic usage**

    When ``user_params = translate = True``, default settings are used.

    >>> example(1)
    ['0:name-of-0']

    Translation may be disabled by passing ``False``, making the helper return immediately.

    >>> example(2, translate=False)
    [0, 1]

    Note that the output type is ``list[int]``, rather than the expected ``list[str]``, in this case.

    Aside from the obvious ``true | false`` behaviour, the helper may also act on the input type.

    **Argument forwarding**

    Dicts are interpreted as keyword-arguments for :meth:`.Translator.translate`.

    >>> example(22, translate={"fmt": "{name} (binary={id:0b})"})[-1]
    'name-of-21 (binary=10101)'

    Plain strings are interpreted as a temporary translation :class:`~id_translation.offline.Format`.

    >>> example(22, translate="{name} (binary={id:0b})")[-1]
    'name-of-21 (binary=10101)'

    This is equivalent to passing ``translate={"fmt": "{name} (binary={id:0b})"}``, as we did above.

    .. note::

       Users may not override the `fixed_params` of the helper instance.

    The helper uses :meth:`.TranslationHelper.convert_user_params` internally, which may also be used to validate the
    configuration.

    **Documenting user arguments**

    Initialized helpers provide methods creating :meth:`user_params <.make_user_params_docstring>`
    and :meth:`type error <.make_type_error_docstring>` docstrings, which may be used as part of the docstring of
    functions that use translation helpers.

    >>> example.__doc__.format(  # doctest: +SKIP
    ...     translate=helper.make_user_params_docstring(),
    ...     type_error=helper.make_type_error_docstring(),
    ... )
    >>> help(example)

    See the :func:`example` function below for output.
"""

import os as _os
import typing as _t
from collections import abc as _abc

from rics.misc import format_kwargs as _format_kwargs
from rics.misc import get_public_module as _get_public_module

from id_translation import Translator as _Translator
from id_translation import translator_typing as _trt
from id_translation import types as _tt
from id_translation.offline import types as _ot
from id_translation.types import TranslatableT

MaximalUntranslatedFractionTypes = int | float
UserParams = (
    bool
    | _ot.FormatType
    | MaximalUntranslatedFractionTypes
    | _trt.TranslateParams[_tt.NameType, _tt.SourceType, _tt.IdType]
)

FactoryFn = _abc.Callable[[], _Translator[_tt.NameType, _tt.SourceType, _tt.IdType]]
FactoryTypes = (
    FactoryFn[_tt.NameType, _tt.SourceType, _tt.IdType] | _Translator[_tt.NameType, _tt.SourceType, _tt.IdType]
)

ALWAYS_RESERVED = ("copy", "translatable")


class TranslationHelper(_t.Generic[_tt.NameType, _tt.SourceType, _tt.IdType]):
    """Helper class for single-purpose translation tasks.

    Args:
        translator_or_factory: A callable ``() -> Translator``, or an initialized :class:`.Translator`.
        user_params_name: Used for reporting errors.
        **fixed_params: Fixed parameters for :meth:`.Translator.translate`. Attempting to override these in
            :meth:`TranslationHelper.apply` will raise an error.

    Notes:
        The https://github.com/rsundqvist/id-translation-project/ template includes functions such as
        :func:`~big_corporation_inc.id_translation.get_singleton`, which are suitable :class:`.Translator` suppliers.
        See the `Big Corporation Inc. <https://rsundqvist.github.io/id-translation-project/>`_ sample docs for more.

    .. seealso:: The :ref:`Typing rules` and :ref:`Examples <translation-helper-examples>` sections.
    """

    def __init__(
        self,
        translator_or_factory: FactoryTypes[_tt.NameType, _tt.SourceType, _tt.IdType],
        user_params_name: str = "translate",
        **fixed_params: _t.Unpack[_trt.TranslateParams[_tt.NameType, _tt.SourceType, _tt.IdType]],
    ) -> None:
        self._user_params_name = user_params_name
        self._get: FactoryFn[_tt.NameType, _tt.SourceType, _tt.IdType]

        if isinstance(translator_or_factory, _Translator):
            self._get = lambda: translator_or_factory
        else:
            self._get = translator_or_factory

        self._fixed = self._validate("fixed_params", fixed_params, protected=self._make())
        self._translated_names: _tt.NameToSource[_tt.NameType, _tt.SourceType] | None = None

    @_t.overload
    def apply(
        self,
        translatable: TranslatableT,
        *,
        user_params: UserParams[_tt.NameType, _tt.SourceType, _tt.IdType],
        copy: _t.Literal[False],
        **default_params: _t.Unpack[_trt.TranslateParams[_tt.NameType, _tt.SourceType, _tt.IdType]],
    ) -> None: ...

    @_t.overload
    def apply(
        self,
        translatable: TranslatableT,
        *,
        copy: _t.Literal[True] = True,
        user_params: _t.Literal[False],
        **default_params: _t.Unpack[_trt.TranslateParams[_tt.NameType, _tt.SourceType, _tt.IdType]],
    ) -> TranslatableT: ...

    @_t.overload
    def apply(
        self,
        translatable: TranslatableT,
        *,
        copy: _t.Literal[True] = True,
        user_params: UserParams[_tt.NameType, _tt.SourceType, _tt.IdType],
        **default_params: _t.Unpack[_trt.TranslateParams[_tt.NameType, _tt.SourceType, _tt.IdType]],
    ) -> _t.Any: ...

    def apply(
        self,
        translatable: TranslatableT,
        *,
        copy: bool = True,
        user_params: UserParams[_tt.NameType, _tt.SourceType, _tt.IdType],
        **default_params: _t.Unpack[_trt.TranslateParams[_tt.NameType, _tt.SourceType, _tt.IdType]],
    ) -> _t.Any | None:
        """Apply translation to `translatable`.

        Keys {always_reserved} are always reserved.

        Args:
            translatable: A data structure to translate.
            copy: If ``False``, translate in-place and return ``None``.
            user_params: {user_params}
            **default_params: Default arguments for the ``translate`` method. May be overridden by `user_params`. If the
                user passes any reserved or fixed keys, a :class:`TypeError` is raised.

        Returns:
            The original `translatable` if `user_params` is ``False``. Otherwise, return a translated copy or
            ``None`` based on the `copy`-setting (see :meth:`.Translator.translate`).

        Raises:
            TypeError: If reserved or fixed keys are passed in the `user_params`.
        """
        return self._apply(translatable, copy=copy, user_params=user_params, default_params=default_params)

    def _apply(
        self,
        translatable: TranslatableT,
        *,
        copy: bool,
        user_params: UserParams[_tt.NameType, _tt.SourceType, _tt.IdType],
        default_params: _trt.TranslateParams[_tt.NameType, _tt.SourceType, _tt.IdType],
    ) -> _t.Any:
        try:
            params = self._process_params(user_params=user_params, default_params=default_params)
        except _AbortTranslation:
            return translatable if copy else None
        _t.assert_type(params, _trt.TranslateParams[_tt.NameType, _tt.SourceType, _tt.IdType])

        translator = self.get_translator()
        result = translator.translate(translatable, copy=copy, **params)  # type: ignore[call-overload]

        self._translated_names = translator.translated_names(with_source=True)
        return result

    def name_to_source(self) -> _tt.NameToSource[_tt.NameType, _tt.SourceType]:
        """Return the name-to-source mapping of the latest :meth:`.apply()`-call."""
        if self._translated_names is None:
            raise ValueError("No names have been translated using this TranslationHelper.")
        return dict(self._translated_names)

    def _process_params(
        self,
        *,
        user_params: UserParams[_tt.NameType, _tt.SourceType, _tt.IdType],
        default_params: _trt.TranslateParams[_tt.NameType, _tt.SourceType, _tt.IdType],
    ) -> _trt.TranslateParams[_tt.NameType, _tt.SourceType, _tt.IdType]:
        self._validate("default_params", default_params)

        if user_params is False:
            # Indicates that no translation should be performed. We could perform this check right away and possibly
            # save some time on validating the default parameters. Reason that we don't is two-fold: 1) It may hide a
            # configuration error, and 2) it is assumed that translation is the most common use case.
            raise _AbortTranslation

        converted_user_params = self.convert_user_params(user_params, validate=False)
        self._validate(self._user_params_name, converted_user_params)

        return {**default_params, **converted_user_params, **self._fixed}

    def convert_user_params(
        self,
        user_params: UserParams[_tt.NameType, _tt.SourceType, _tt.IdType],
        validate: bool = True,
    ) -> _trt.TranslateParams[_tt.NameType, _tt.SourceType, _tt.IdType]:
        """Convert user parameters.

        Args:
            user_params: End-user parameters.
            validate: If ``False``, skip the regular fixed parameter validation.

        Returns:
            Valid :meth:`.Translator.translate` parameters
        """
        if user_params is False:
            msg = f"Cannot convert {self._user_params_name}=False."
            raise TypeError(msg)

        if user_params is True:
            return self._make()

        if isinstance(user_params, str):
            return self._make(fmt=user_params)

        if isinstance(user_params, _t.get_args(MaximalUntranslatedFractionTypes)):
            return self._make(max_fails=user_params)

        if isinstance(user_params, dict):
            params = self._make(**user_params)
            return self._validate(self._user_params_name, params) if validate else params

        types = (typ.__name__ for typ in (bool, str, float, dict))
        msg = f"type({self._user_params_name}) is {type(user_params).__name__}. Expected: ({', '.join(types)})."
        raise TypeError(msg)

    @classmethod
    def _make(
        cls, **params: _t.Unpack[_trt.TranslateParams[_tt.NameType, _tt.SourceType, _tt.IdType]]
    ) -> _trt.TranslateParams[_tt.NameType, _tt.SourceType, _tt.IdType]:
        """Convenience function to avoid having to repeat the type variables."""
        return params

    def get_translator(self) -> _Translator[_tt.NameType, _tt.SourceType, _tt.IdType]:
        """Return a ``Translator`` instance."""
        return self._get()

    def __repr__(self) -> str:
        parts = [_get_public_module(self._get, include_name=True, resolve_reexport=True)]

        if self._fixed:
            parts.append(_format_kwargs(self._fixed))

        if (user_param_name := self._user_params_name) != "translate":
            parts.append(f"{user_param_name=}")

        return f"{type(self).__name__}({', '.join(parts)})"

    def _validate(
        self,
        name: str,
        params: _trt.TranslateParams[_tt.NameType, _tt.SourceType, _tt.IdType],
        *,
        protected: _trt.TranslateParams[_tt.NameType, _tt.SourceType, _tt.IdType] | None = None,
    ) -> _trt.TranslateParams[_tt.NameType, _tt.SourceType, _tt.IdType]:
        if keys := {*(self._fixed if protected is None else protected), *ALWAYS_RESERVED}.intersection(params):
            msg = f"Found protected {keys=} in {name}={params}."
            raise TypeError(msg)

        return params

    def make_user_params_docstring(self) -> str:
        """Create description for `user_params`.

        Example output below.

        Args:
            user_params: {user_params}

        Output may vary depending on helper settings.
        """
        reserved = set(self._fixed)

        parts = [
            "Translation options. Set to ``False`` to disable.",
            "If :class:`dict`, use as keyword-arguments for :attr:`.Translator.translate` (raises"
            f" :py:class:`TypeError` for {len(reserved) + len(ALWAYS_RESERVED)} reserved keys).",
        ]

        types = [
            (str, "fmt", "see :class:`.Format`"),
            (float, "max_fails", "where 1=disable check, 0=no missing IDs allowed"),
        ]

        type_parts = []
        for typ, key, hint in types:
            if key in reserved:
                continue

            template = ":class:`{.__name__}` = ``{!r}`` ({})"
            type_parts.append(template.format(typ, key, hint))

        if type_parts:
            parts.append("Other types:")
            parts.append(", ".join(type_parts) + ".")

        return " ".join(parts)

    def make_type_error_docstring(self) -> str:
        """Create description for ``TypeError``.

        Example output below.

        Raises:
            TypeError: {type_error}

        Output may vary depending on helper settings.
        """
        reserved = *ALWAYS_RESERVED, *self._fixed

        parts = (
            f"Raised if `{self._user_params_name}` is a ``dict`` containing any of the {len(reserved)} the reserved keys: ",
            ", ".join(f"``'{key}'``" for key in reserved),
            ".",
        )
        return "".join(parts)

    def make_docstrings(
        self,
        *,
        user_params_key: str | None = None,
        type_error_key: str = "type_error",
    ) -> dict[str, str]:
        """Convenience method for creating multiple docstrings.

        Args:
            user_params_key: Key for :meth:`make_user_params_docstring` output. Default is `user_params_name`.
            type_error_key: Key for :meth:`make_type_error_docstring` output.

        Returns:
            A dict of docstrings.
        """
        return {
            self._user_params_name if user_params_key is None else user_params_key: self.make_user_params_docstring(),
            type_error_key: self.make_type_error_docstring(),
        }


class _AbortTranslation(Exception):  # noqa: N818
    pass


if __doc__:

    def _patch_docstrings() -> None:
        cls = TranslationHelper[str, str, str]
        dummy = cls(_Translator, user_params_name="<user_params_name>")
        docstrings = {
            "always_reserved": ", ".join(f"``'{key}'``" for key in ALWAYS_RESERVED),
            **dummy.make_docstrings(user_params_key="user_params"),
        }

        for func in cls.apply, cls.make_user_params_docstring, cls.make_type_error_docstring:
            assert func.__doc__, "missing docstring"  # noqa S101
            func.__doc__ = func.__doc__.format_map(docstrings)

    _patch_docstrings()
    del _patch_docstrings

    if _os.environ.get("SPHINX_BUILD") == "true":  # pragma: no cover
        helper = TranslationHelper[str, str, int](_Translator, "translate", names="name")

        def example(
            n: int,
            *,
            translate: UserParams[str, str, int] = True,
        ) -> list[str]:
            """Create and translate the first `n` integers.

            Docstrings for `translate` and ``TypeError`` were produced by :meth:`~.TranslationHelper.make_docstrings`.

            Args:
                n: Number of integers to create.
                translate: {translate}

            Raises:
                TypeError: {type_error}

            Returns:
                A list.
            """
            items: list[str] = helper.apply(
                list(range(n)),
                copy=True,
                user_params=translate,
                fmt="{id}:{name}",
            )
            return items

        example.__doc__ = example.__doc__.format(  # type: ignore[union-attr]
            translate=helper.make_user_params_docstring(),
            type_error=helper.make_type_error_docstring(),
        )



"""Handler which pretty-prints (color) logger output to stdout.

Used by :func:`.enable_verbose_debug_messages`.
"""

import logging
from datetime import datetime
from enum import StrEnum
from random import shuffle
from re import Match, compile
from threading import Lock
from types import TracebackType
from typing import TypeAlias

ExcInfo: TypeAlias = tuple[type[BaseException], BaseException, TracebackType | None] | tuple[None, None, None]


class Color(StrEnum):
    """Colors used for highlighting."""

    reset = "\033[0m"
    red = "\033[31m"
    green = "\033[32m"
    yellow = "\033[33m"
    blue = "\033[34m"
    magenta = "\033[35m"
    cyan = "\033[36m"
    gray = "\033[37m"
    orange = "\033[93m"


LEVEL_TO_COLOR: dict[int, Color] = {
    logging.DEBUG: Color.magenta,
    logging.INFO: Color.green,
    logging.WARNING: Color.orange,
    logging.ERROR: Color.red,
}

PATTERN_TO_COLOR: tuple[tuple[str, Color], ...] = (
    # Strings
    (r"'.*?'", Color.green),
    (r'".*?"', Color.green),
    # Numerics
    (r"(?:0x)[abcdef\d]+", Color.magenta),  # Hex string
    (r"(?:0b)[01]+", Color.magenta),  # Bin string
    (r"[+-]?\d+(?:.\d+)?", Color.magenta),  # Numbers; includes things like 3.2 or 3x2.
    (r"[+-]?(?:NaN|inf|∞)", Color.magenta),  # Infinity/NaN
    # Misc
    ("None|True|False", Color.gray),  # Constants
    (r"\w+=", Color.red),  # Keyword argument
)


class DebugLoggingFormatter(logging.Formatter):
    """Helper class used to display debug out.

    Args:
        less: If ``True``, abbreviate output.
        indent_style: Indentation prefix. Empty=disabled.
    """

    def __init__(
        self,
        *,
        less: bool = False,
        indent_style: str = "  ",
    ) -> None:
        super().__init__()

        self._less = less

        pattern_to_color = PATTERN_TO_COLOR
        if less:
            pattern_to_color = tuple((p, c) for p, c in pattern_to_color if c == Color.magenta)
        self._pattern_to_color = pattern_to_color
        self._pattern = compile("|".join(f"({pattern})" for i, (pattern, _) in enumerate(pattern_to_color)))

        self._indent_style = indent_style
        self._indent: dict[str, int] = {}

        self._task_id_styles: list[tuple[str, Color]] = [
            ("🦋", Color.yellow),
            ("🍏", Color.green),
            ("🫚", Color.orange),
            ("🎷", Color.yellow),
            ("🎃", Color.orange),
            ("🦄", Color.magenta),
            ("🫎", Color.gray),
            ("🪻", Color.magenta),
            ("🐳", Color.blue),
        ]

        shuffle(self._task_id_styles)
        self._task_id_style: dict[int, tuple[str, Color]] = {}

        self._lock = Lock()

    def formatMessage(self, record: logging.LogRecord) -> str:  # noqa: N802
        """Emit pretty-printed log record."""
        task_id = self._task_id(record)
        created = self._created(record)
        indentation = self._indentation(record)
        level = self._level(record)
        logger = self._logger(record)
        message = self._message(record)
        parts = [task_id, created, indentation, level, logger, message]

        return " ".join(filter(bool, parts))

    def formatException(self, ei: ExcInfo) -> str:  # noqa: N802
        """Emit exception with highlighting."""
        s = super().formatException(ei)
        s = self._highlight(s)
        return s

    def _indentation(self, record: logging.LogRecord) -> str:
        task_id = getattr(record, "task_id", None)
        if task_id is None or self._less or self._indent_style == "":
            return ""

        stage = ""
        if event_key := getattr(record, "event_key", ""):
            stage = event_key.rpartition(":")[2]

        self._indent.setdefault(task_id, 0)  # Just in case; enter should come first.

        if stage == "exit":
            self._indent[task_id] -= 1

        indentation = self._indent_style * self._indent[task_id]

        if stage == "enter":
            self._indent[task_id] += 1

        return indentation

    def _task_id(self, record: logging.LogRecord) -> str:
        task_id = getattr(record, "task_id", None)
        if task_id is None:
            return "[-- ------]"

        with self._lock:
            symbol, color = self._select_task_id_style(task_id)

        return f"[{color}{symbol} {hex(task_id):0<6}{Color.reset}]"

    def _select_task_id_style(self, task_id: int) -> tuple[str, Color]:
        if task_id in self._task_id_style:
            return self._task_id_style[task_id]

        index = len(self._task_id_style) % len(self._task_id_styles)
        symbol = self._task_id_styles[index]
        self._task_id_style[task_id] = symbol
        return symbol

    def _created(self, record: logging.LogRecord) -> str:
        date = "" if self._less else f"{Color.gray}%d %b{Color.reset} "
        time = f"{Color.orange}%H:%M:%S{Color.yellow}.%f{Color.reset}"
        created = datetime.fromtimestamp(record.created)
        return "[" + created.strftime(f"{date}{time}") + "]"

    def _level(self, record: logging.LogRecord) -> str:
        if self._less:
            return f"[{record.levelname}]"
        color = LEVEL_TO_COLOR.get(record.levelno, Color.reset)
        return f"[{color}{record.levelname}{Color.reset}]"

    @classmethod
    def _logger(cls, record: logging.LogRecord) -> str:
        if event_key := getattr(record, "event_key", ""):
            meth, _, stage = event_key.rpartition(":")
            stage = {"enter": "🚀", "exit": "✅"}.get(stage, stage)
            name = f"{stage} {Color.blue}{meth}{Color.reset}"
        else:
            name = Color.gray + record.name + Color.reset

        return f"[{name}]"

    def _message(self, record: logging.LogRecord) -> str:
        message = record.getMessage()
        return self._highlight(message)

    def _highlight(self, s: str) -> str:
        return self._pattern.sub(self._highlight_match, s)

    def _highlight_match(self, match: Match[str]) -> str:
        assert match.lastindex is not None  # noqa: S101
        color = self._pattern_to_color[match.lastindex - 1][1]
        return color + match.group(0) + color.reset



"""Default factory implementations.

These may be used as baselines when overwriting ``FACTORY``-attributes in the :class:`.TranslatorFactory` class.
"""

from ._cache_access import default_cache_access_factory
from ._fetcher import default_fetcher_factory
from ._mapper import default_mapper_factory
from ._transformer import default_transformer_factory

__all__ = [
    "default_cache_access_factory",
    "default_fetcher_factory",
    "default_mapper_factory",
    "default_transformer_factory",
]



"""Insertion and extraction of IDs and translations."""

from abc import abstractmethod
from collections.abc import Sequence
from typing import Any, ClassVar, Generic

from ..offline import TranslationMap
from ..types import IdType, NameType, SourceType, TranslatableT


class DataStructureIO(Generic[TranslatableT, NameType, SourceType, IdType]):
    """Insertion and extraction of IDs and translations."""

    priority: ClassVar[int] = 10_000
    """Determines order in which IOs are considered (higher = earlier).

    Set `priority < 0` to disable.
    """

    @classmethod
    def register(cls) -> None:
        """Register this implementation for all :class:`.Translator` instances.

        See :func:`.dio.register_io` for details.
        """
        from ._resolve import register_io  # noqa: PLC0415

        return register_io(cls)

    @classmethod
    def is_registered(cls) -> bool:
        """Returns registration status for this implementation.

        See :func:`.dio.is_registered` for details.
        """
        from ._resolve import is_registered  # noqa: PLC0415

        return is_registered(cls)

    @classmethod
    def get_rank(cls) -> int:
        """Return the rank of this implementation.

        See :func:`.dio.get_resolution_order` for details. The highest possible rank is 0.

        Returns:
            Implementation rank.

        Raises:
            ValueError: If the implementation is not registered.
        """
        from ._resolve import get_resolution_order  # noqa: PLC0415

        try:
            return get_resolution_order(real=True).index(cls)
        except ValueError:
            msg = (
                f"Not registered: {cls.__name__}"
                f"\nHint: Use {cls.register.__qualname__}() to register this implementation."
                "\nHint: https://id-translation.readthedocs.io/en/stable/api/id_translation.dio.html#user-defined-integrations"
            )
            raise ValueError(msg) from None

    @classmethod
    @abstractmethod
    def handles_type(cls, arg: Any) -> bool:
        """Return ``True`` if the implementation handles data for the type of `arg`."""

    def names(self, translatable: TranslatableT) -> list[NameType] | None:
        """Extract names from `translatable`.

        Args:
            translatable: Data to extract names from.

        Returns:
            A list of names to translate. Returns ``None`` if names cannot be extracted.
        """
        return translatable.name if hasattr(translatable, "name") else None

    @abstractmethod
    def extract(self, translatable: TranslatableT, names: list[NameType]) -> dict[NameType, Sequence[IdType]]:
        """Extract IDs from `translatable`.

        Args:
            translatable: Data to extract IDs from.
            names: List of names in `translatable` to extract IDs for.

        Returns:
            A dict ``{name: ids}``.
        """

    @abstractmethod
    def insert(
        self,
        translatable: TranslatableT,  # TODO Higher-Kinded TypeVars
        names: list[NameType],
        tmap: TranslationMap[NameType, SourceType, IdType],
        copy: bool,
    ) -> Any | None:
        """Insert translations into `translatable`.

        Args:
            translatable: Data to translate. Modified iff ``copy=False``.
            names: Names in `translatable` to translate.
            tmap: Translations for IDs in `translatable`.
            copy: If ``True``, modify contents of the original `translatable`. Otherwise, returns a copy.

        Returns:
            A copy of `translatable` if ``copy=True``, ``None`` otherwise.

        Raises:
            NotInplaceTranslatableError: If ``copy=False`` for a type which is not translatable in-place.
        """



import logging
import warnings
from collections.abc import Iterable, Mapping
from concurrent.futures import Future, ThreadPoolExecutor, as_completed
from copy import deepcopy
from time import perf_counter
from typing import Any, Literal, Never, Self

from rics.collections.dicts import reverse_dict
from rics.logs import LogLevel, convert_log_level
from rics.misc import tname
from rics.strings import format_seconds as fmt_sec
from rics.types import LiteralHelper

from ..logging import generate_task_id, get_event_key
from ..offline.types import SourcePlaceholderTranslations
from ..types import IdType, SourceType
from . import AbstractFetcher, Fetcher, exceptions
from .types import IdsToFetch, Operation

LOGGER = logging.getLogger(__package__).getChild("MultiFetcher")

FetchResult = tuple[int, SourcePlaceholderTranslations[SourceType]]

OnSourceConflict = Literal["raise", "warn", "ignore"]


class MultiFetcher(Fetcher[SourceType, IdType]):
    """Fetcher which combines the results of other fetchers.

    Args:
        *children: Fetchers to wrap.
        max_workers: Number of threads to use for fetching. Fetch instructions will be dispatched using a
             :py:class:`~concurrent.futures.ThreadPoolExecutor`. Individual fetchers will be called at most once per
             ``fetch()`` or ``fetch_all()`` call made with the ``MultiFetcher``.
        on_source_conflict: Action to take when multiple fetchers :meth:`claim <.Fetcher.initialize_sources>` the same source.
        fetcher_discarded_log_level: Level used when discarding :attr:`~.Fetcher.optional` fetchers.
    """

    def __init__(
        self,
        *children: Fetcher[SourceType, IdType],
        max_workers: int = 1,
        on_source_conflict: OnSourceConflict = "raise",
        fetcher_discarded_log_level: LogLevel = "DEBUG",
    ) -> None:
        for pos, f in enumerate(children):
            if not isinstance(f, Fetcher):  # pragma: no cover
                raise TypeError(f"Argument {pos} is of type {type(f)}, expected Fetcher subtype.")

        self._id_to_rank: dict[int, int] = {id(f): rank for rank, f in enumerate(children)}
        self._id_to_fetcher: dict[int, Fetcher[SourceType, IdType]] = {id(f): f for f in children}
        self.max_workers: int = max_workers

        self._on_source_conflict = OSC_HELPER.check(on_source_conflict)
        self._discard_level = convert_log_level(fetcher_discarded_log_level, name="fetcher_discarded_log_level")

        if len(self._id_to_rank) != len(children):
            raise ValueError("Repeat fetcher instance(s)!")  # pragma: no cover

        self._placeholders: dict[SourceType, list[str]] | None = None
        self._source_to_id: dict[SourceType, int] = {}

    @property
    def allow_fetch_all(self) -> bool:
        return all(f.allow_fetch_all for f in self._id_to_fetcher.values())  # pragma: no cover

    @property
    def online(self) -> bool:
        return all(f.online for f in self._id_to_fetcher.values())  # pragma: no cover

    def close(self) -> None:
        """Close all :attr:`child <children>` fetchers."""
        for fetcher in self.children:
            fetcher.close()

    @property
    def children(self) -> list[Fetcher[SourceType, IdType]]:
        """Return child fetchers sorted by rank."""
        self.initialize_sources()
        children = [*self._id_to_fetcher.values()]
        # children.sort(key=lambda fetcher: self._id_to_rank[id(fetcher)])
        return children

    def get_child(self, source: SourceType) -> Fetcher[SourceType, IdType]:
        """Return child fetcher for the given source."""
        self.initialize_sources()

        child_id = self._source_to_id[source]
        fetcher = self._id_to_fetcher[child_id]
        return fetcher

    def get_sources(self, child: Fetcher[SourceType, IdType] | int) -> list[SourceType]:
        """Return sources for the given child."""
        if not isinstance(child, int):
            child = id(child)
        self.initialize_sources()
        return [source for source, child_id in self._source_to_id.items() if child_id == child]

    @property
    def placeholders(self) -> dict[SourceType, list[str]]:
        if self._placeholders is None:
            self.initialize_sources()
            return self.placeholders

        return self._placeholders

    def initialize_sources(self, task_id: int | None = None, *, force: bool = False) -> None:
        """Perform source discovery.

        Perform source discovery for all :attr:`children`, discarding :attr:`optional <.Fetcher.optional>` children that
        raise or do not return any sources when their respective :meth:`.Fetcher.initialize_sources` methods are
        called.

        Args:
            task_id: Used for logging.
            force: If ``True``, perform full discovery even if sources are already known.

        See Also:
            🔑 This is a key event method. See :ref:`key-events` for details.

        Notes:
            Calling this method multiple times will not recover previously discarded optional child fetchers.
        """
        if not (self._placeholders is None or force):
            return

        start = perf_counter()
        if task_id is None:
            task_id = generate_task_id()

        LOGGER.debug(
            "Begin initialization of %i children.",
            len(self._id_to_fetcher),
            extra={"task_id": task_id, "event_key": get_event_key(self.initialize_sources, "enter")},
        )

        fid_to_placeholders = self._initialize_sources(task_id)
        self._source_to_id = self._make_source_to_id(fid_to_placeholders, task_id)

        self._placeholders = {}
        for fid, source_to_placeholders in fid_to_placeholders.items():
            discarded = dict(source_to_placeholders)
            placeholders = {
                source: placeholders
                for source, placeholders in source_to_placeholders.items()
                if self._source_to_id[source] == fid
            }

            if placeholders:
                self._placeholders.update(placeholders)
                continue

            self._handle_all_sources_outranked(task_id, fetcher_id=fid, discarded=discarded)

        if not self._id_to_fetcher:
            warnings.warn("No fetchers. See log output for more information.", UserWarning, stacklevel=1)

        if LOGGER.isEnabledFor(logging.DEBUG):
            seconds = perf_counter() - start
            event_key = get_event_key(self.initialize_sources, "exit")
            LOGGER.debug(
                f"Finished initialization {len(self._id_to_fetcher)} children and "
                f"{len(self._source_to_id)} sources in {fmt_sec(seconds)}.",
                extra={"task_id": task_id, "seconds": seconds, "event_key": event_key},
            )

    def _initialize_sources(self, task_id: int) -> dict[int, dict[SourceType, list[str]]]:
        retval: dict[int, dict[SourceType, list[str]]] = {}

        log_level = self._discard_level

        for fid, fetcher in list(self._id_to_fetcher.items()):
            if fetcher.optional:
                try:
                    fetcher.initialize_sources(task_id, force=True)
                    placeholders = fetcher.placeholders
                except Exception as e:
                    LOGGER.log(
                        log_level,
                        "Discarding optional %s: Raised\n    %s\nwhen getting sources.",
                        self.format_child(fid),
                        f"{type(e).__name__}: {e}",
                        exc_info=True,
                    )
                    fetcher.close()
                    del self._id_to_rank[fid]
                    del self._id_to_fetcher[fid]
                    continue

                if len(placeholders) == 0:
                    if LOGGER.isEnabledFor(log_level):
                        LOGGER.log(
                            log_level,
                            f"Discarding optional {self.format_child(fetcher)}: No sources.",
                            extra={"task_id": task_id},
                        )

                    fetcher.close()
                    del self._id_to_rank[fid]
                    del self._id_to_fetcher[fid]
                    continue

            else:
                try:
                    fetcher.initialize_sources(task_id, force=True)
                except Exception as e:
                    self._raise_with_notes(e, fetcher)

                placeholders = fetcher.placeholders

                if len(placeholders) == 0 and LOGGER.isEnabledFor(logging.WARNING):
                    LOGGER.warning(
                        f"Required {self.format_child(fetcher)} does not provide any sources.",
                        extra={"task_id": task_id},
                    )

            retval[fid] = placeholders

        return retval

    def _make_source_to_id(
        self,
        fid_to_placeholders: Mapping[int, Iterable[SourceType]],
        task_id: int,
    ) -> dict[SourceType, int]:
        if not fid_to_placeholders:
            return {}

        source_ranks: dict[SourceType, int] = {}
        retval: dict[SourceType, int] = {}

        for fid, sources in fid_to_placeholders.items():
            rank = self._id_to_rank[fid]
            for source in sources:
                if source in retval:
                    self._log_rejection(source, rank, source_ranks[source], "INITIALIZE_SOURCES", task_id)
                else:
                    retval[source] = fid
                    source_ranks[source] = rank

        return retval

    def fetch(
        self,
        ids_to_fetch: Iterable[IdsToFetch[SourceType, IdType]],
        placeholders: Iterable[str] = (),
        *,
        required: Iterable[str] = (),
        task_id: int | None = None,
        enable_uuid_heuristics: bool = False,
    ) -> SourcePlaceholderTranslations[SourceType]:
        if task_id is None:
            task_id = generate_task_id()
        self.initialize_sources(task_id)

        tasks: dict[int, list[IdsToFetch[SourceType, IdType]]] = {}
        sources = []
        unknown_sources = []
        for idt in ids_to_fetch:
            source = idt.source
            child_id = self._source_to_id.get(source)
            if child_id is None:
                unknown_sources.append(source)
                continue
            tasks.setdefault(child_id, []).append(idt)
            sources.append(source)

        if unknown_sources:
            raise exceptions.UnknownSourceError(unknown_sources, self.sources)

        placeholders = tuple(placeholders)
        required = tuple(required)

        n_sources_and_fetchers = f"{len(sources)} sources using {len(tasks)} different fetchers"

        start = perf_counter()
        if LOGGER.isEnabledFor(logging.DEBUG):
            LOGGER.debug(
                f"Dispatch FETCH jobs for {n_sources_and_fetchers} on {self.max_workers} threads.",
                extra=dict(
                    task_id=task_id,
                    event_key=get_event_key(self.fetch, "enter"),
                    sources=sources,
                    placeholders=placeholders,
                    required_placeholders=required,
                    max_workers=self.max_workers,
                    num_fetchers=len(tasks),
                    fetch_all=False,
                ),
            )

        def fetch(fid: int) -> FetchResult[SourceType]:
            fetcher = self._id_to_fetcher[fid]
            if LOGGER.isEnabledFor(logging.DEBUG):
                LOGGER.debug(
                    f"Begin FETCH job for {len(tasks[fid])} sources using {self.format_child(fetcher)}.",
                    extra={"task_id": task_id},
                )

            try:
                result = fetcher.fetch(
                    tasks[fid],
                    placeholders,
                    required=required,
                    task_id=task_id,
                    enable_uuid_heuristics=enable_uuid_heuristics,
                )
            except Exception as e:
                self._raise_with_notes(e, fetcher)
            return fid, result

        with ThreadPoolExecutor(max_workers=self.max_workers, thread_name_prefix=tname(self)) as executor:
            futures = [executor.submit(fetch, fid) for fid in tasks]
            ans = self._gather(futures, operation="FETCH", task_id=task_id)

        if LOGGER.isEnabledFor(logging.DEBUG):
            seconds = perf_counter() - start
            LOGGER.debug(
                f"Completed FETCH jobs for {n_sources_and_fetchers} in {fmt_sec(seconds)}.",
                extra=dict(
                    task_id=task_id,
                    event_key=get_event_key(self.fetch, "exit"),
                    seconds=seconds,
                    sources=len(ans),
                    max_workers=self.max_workers,
                    num_fetchers=len(tasks),
                    fetch_all=False,
                ),
            )

        return ans

    def fetch_all(
        self,
        placeholders: Iterable[str] = (),
        *,
        required: Iterable[str] = (),
        sources: set[SourceType] | None = None,
        task_id: int | None = None,
        enable_uuid_heuristics: bool = False,
    ) -> SourcePlaceholderTranslations[SourceType]:
        if task_id is None:
            task_id = generate_task_id()
        self.initialize_sources(task_id)

        placeholders = tuple(placeholders)
        required = tuple(required)

        start = perf_counter()
        if LOGGER.isEnabledFor(logging.DEBUG):
            LOGGER.debug(
                f"Dispatch FETCH_ALL jobs for {len(self.children)} fetchers on {self.max_workers} threads.",
                extra=dict(
                    task_id=task_id,
                    event_key=get_event_key(self.fetch_all, "enter"),
                    placeholders=placeholders,
                    required_placeholders=required,
                    sources=None if sources is None else [*sources],
                    max_workers=self.max_workers,
                    num_fetchers=len(self.children),
                    fetch_all=True,
                ),
            )

        def fetch_all(fetcher: Fetcher[SourceType, IdType]) -> FetchResult[SourceType]:
            result = fetcher.fetch_all(
                placeholders,
                required=required,
                sources=None if sources is None else sources.intersection(fetcher.sources),
                task_id=task_id,
                enable_uuid_heuristics=enable_uuid_heuristics,
            )
            return id(fetcher), result

        children = self.children if sources is None else [c for c in self.children if sources.intersection(c.sources)]
        with ThreadPoolExecutor(max_workers=self.max_workers, thread_name_prefix=tname(self)) as executor:
            futures = [executor.submit(fetch_all, fetcher) for fetcher in children]
            ans = self._gather(futures, operation="FETCH_ALL", task_id=task_id)

        if LOGGER.isEnabledFor(logging.DEBUG):
            seconds = perf_counter() - start
            LOGGER.debug(
                f"Completed FETCH_ALL jobs for {len(ans)} sources using "
                f"{len(self.children)} fetchers in {fmt_sec(seconds)}.",
                extra=dict(
                    task_id=task_id,
                    event_key=get_event_key(self.fetch_all, "exit"),
                    seconds=seconds,
                    sources=list(ans),
                    max_workers=self.max_workers,
                    num_fetchers=len(self.children),
                    fetch_all=True,
                ),
            )
        return ans

    @property
    def on_source_conflict(self) -> OnSourceConflict:
        """Action to take when multiple fetchers :meth:`claim <.Fetcher.initialize_sources>` the same source."""
        return self._on_source_conflict

    def _gather(
        self,
        futures: Iterable[Future[FetchResult[SourceType]]],
        operation: Operation,
        task_id: int,
    ) -> SourcePlaceholderTranslations[SourceType]:
        ans: SourcePlaceholderTranslations[SourceType] = {}
        source_ranks: dict[SourceType, int] = {}

        for future in as_completed(futures):
            fid, translations = future.result()
            rank = self._id_to_rank[fid]
            self._process_future_result(translations, rank, source_ranks, ans, operation, task_id)
        return ans

    def _process_future_result(
        self,
        translations: SourcePlaceholderTranslations[SourceType],
        rank: int,
        source_ranks: dict[SourceType, int],
        ans: SourcePlaceholderTranslations[SourceType],
        operation: Operation,
        task_id: int,
    ) -> None:
        for source_translations in translations.values():
            source = source_translations.source
            other_rank = source_ranks.setdefault(source, rank)

            if other_rank != rank:
                self._log_rejection(source, rank, other_rank, operation, task_id)
                if rank > other_rank:
                    continue  # Don't save -- other rank is greater (lower-is-better).

            ans[source] = source_translations

    def _log_rejection(self, source: SourceType, rank0: int, rank1: int, operation: Operation, task_id: int) -> None:
        accepted_rank, rejected_rank = (rank0, rank1) if rank0 < rank1 else (rank1, rank0)

        rank_to_id = reverse_dict(self._id_to_rank)
        accepted = self.format_child(rank_to_id[accepted_rank])
        rejected = self.format_child(rank_to_id[rejected_rank])

        hints = []
        if operation == "INITIALIZE_SOURCES":
            msg = f"Discarded {source=} retrieved from {rejected} since the {accepted} already claimed same source."
            hints.append("Hint: Rank is determined input order at initialization.")
            on_source_conflict = self.on_source_conflict
        elif operation == "FETCH_ALL":
            msg = (
                f"Dropping translations for {source=} returned by the {rejected} since {operation=}."
                f" Will use {accepted} translations instead."
            )
            on_source_conflict = "ignore"
        else:  # Bad Fetcher.fetch implementation; should be rare.
            fetcher = self._id_to_fetcher[rank_to_id[rejected_rank]]
            cls = tname(fetcher, include_module=True)
            msg = f"Dropping translations for {source=} returned by the {rejected}; this source belongs to the {accepted}."
            hints.append(f"Hint: The implementation of {cls} may be incorrect.")
            on_source_conflict = "warn"

        extra = {"task_id": task_id, "source": source}
        if on_source_conflict == "raise":
            LOGGER.error(msg, extra=extra)
            exc = exceptions.DuplicateSourceError(msg)
            for hint in hints:
                exc.add_note(hint)
            raise exc
        if on_source_conflict == "warn":
            LOGGER.warning(msg, extra=extra)

            warnings.warn(msg, exceptions.DuplicateSourceWarning, stacklevel=3)
            msg += "\n".join(hints)
        elif on_source_conflict == "ignore" and LOGGER.isEnabledFor(logging.DEBUG):
            LOGGER.debug(msg, extra=extra)

    def __repr__(self) -> str:
        max_workers = self.max_workers
        fetchers = "\n    ".join(f"{f}," for f in self._id_to_fetcher.values())
        return f"{tname(self)}({max_workers=}, fetchers=[\n    {fetchers}\n])"

    def format_child(self, fetcher: int | Fetcher[SourceType, IdType]) -> str:
        """Format a managed fetcher with rank and hex ID."""
        if isinstance(fetcher, int):
            fetcher = self._id_to_fetcher[fetcher]
        fetcher_id = id(fetcher)
        rank = self._id_to_rank[fetcher_id]
        return f"rank-{rank} fetcher {fetcher} at {hex(fetcher_id)}"

    def _handle_all_sources_outranked(
        self, task_id: int, *, fetcher_id: int, discarded: dict[SourceType, list[str]]
    ) -> None:
        fetcher = self._id_to_fetcher[fetcher_id]

        if LOGGER.isEnabledFor(logging.WARNING) and fetcher.sources:
            reason = "All sources found in higher-ranking fetchers"
            pretty = self.format_child(fetcher)
            LOGGER.warning(
                f"Discarding optional {pretty}: {reason}."
                if fetcher.optional
                else f"Required {pretty} is useless, but will be kept: {reason}.",
                extra={"task_id": task_id, "discarded": discarded, "fetcher_id": fetcher_id},
            )

        if fetcher.optional:
            fetcher.close()
            del self._id_to_rank[fetcher_id]
            del self._id_to_fetcher[fetcher_id]

    def _raise_with_notes(self, e: BaseException, fetcher: Fetcher[SourceType, IdType]) -> Never:
        note = f"Context (added by {type(self).__name__}):"

        # Add config file. Mirrors logic used in the abstract fetcher.
        if isinstance(fetcher, AbstractFetcher):
            for idx in fetcher.identifiers:
                if idx.endswith("toml"):
                    note += f"\n -  file= '{idx}'"
                    break

        note += f"\n - child= {self.format_child(fetcher)}"
        e.add_note(note)
        raise e

    def __deepcopy__(self, memo: dict[int, Any] = {}) -> Self:  # noqa: B006
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result

        dicts = self._copy_dicts(memo)
        for k, v in self.__dict__.items():
            setattr(result, k, dicts[k] if k in dicts else deepcopy(v, memo))

        return result

    def _copy_dicts(self, memo: dict[int, Any]) -> dict[str, dict[str, Any] | dict[int, Any]]:
        new_id_to_fetcher: dict[int, Fetcher[SourceType, IdType]] = {}
        old_id_to_new_id: dict[int, int] = {}

        members = self.__dict__

        for old_id, old_fetcher in members["_id_to_fetcher"].items():
            try:
                new_fetcher = deepcopy(old_fetcher, memo)
            except TypeError as e:
                new_fetcher = old_fetcher

                # This hides the Translator.copy(fetcher=Translator.fetcher) warning emitted in the caller!
                fetcher_cls = type(old_fetcher).__name__
                msg = f"deepcopy() failed ({type(e).__name__}: {e}). Reusing {self.format_child(old_fetcher)}"
                LOGGER.warning(msg, exc_info=True, extra={"fetcher_class": fetcher_cls})

            new_id = id(new_fetcher)
            new_id_to_fetcher[new_id] = new_fetcher
            old_id_to_new_id[old_id] = new_id

        return {
            "_id_to_fetcher": new_id_to_fetcher,
            "_id_to_rank": {old_id_to_new_id[old_id]: rank for old_id, rank in members["_id_to_rank"].items()},
            "_source_to_id": {source: old_id_to_new_id[old_id] for source, old_id in members["_source_to_id"].items()},
        }


OSC_HELPER: LiteralHelper[OnSourceConflict] = LiteralHelper(
    OnSourceConflict,
    default_name="on_source_conflict",
    type_name="OnSourceConflict",
)



from collections.abc import Mapping, Sequence
from typing import TYPE_CHECKING, Any, Unpack

from ..offline.types import PlaceholderTranslations, SourcePlaceholderTranslations
from ..translator_typing import AbstractFetcherParams
from ..types import ID, IdType, SourceType
from ._abstract_fetcher import AbstractFetcher
from .types import FetchInstruction

if TYPE_CHECKING:
    import pandas


class MemoryFetcher(AbstractFetcher[SourceType, IdType]):
    """Fetch from memory.

    This is essentially a thin wrapper for the :class:`.PlaceholderTranslations` class.

    Args:
        data: A dict ``{source: PlaceholderTranslations}`` to fetch from.
        return_all: If ``False``, return only the requested IDs and placeholders.
        **kwargs: See :class:`.AbstractFetcher`.
    """

    def __init__(
        self,
        data: (
            SourcePlaceholderTranslations[SourceType]
            | Mapping[SourceType, PlaceholderTranslations[SourceType]]
            | Mapping[SourceType, "pandas.DataFrame"]
            | Mapping[SourceType, Mapping[str, Sequence[Any]]]
        ),
        return_all: bool = True,
        **kwargs: Unpack[AbstractFetcherParams[SourceType, IdType]],
    ) -> None:
        super().__init__(**kwargs)
        self._data = {
            source: pht
            if isinstance(pht, PlaceholderTranslations)
            else PlaceholderTranslations[SourceType].make(source, pht)
            for source, pht in data.items()
        }
        self._return_all = return_all

    @property
    def return_all(self) -> bool:
        """If ``True``, :meth:`fetch_translations` will filter by ID."""
        return self._return_all

    def _initialize_sources(self, task_id: int) -> dict[SourceType, list[str]]:  # noqa: ARG002
        return {source: list(pht.placeholders) for source, pht in self._data.items()}

    def fetch_translations(self, instr: FetchInstruction[SourceType, IdType]) -> PlaceholderTranslations[SourceType]:
        ret = self._data[instr.source]

        if self.return_all:
            return ret

        placeholder_indices = [
            ret.placeholders.index(placeholder)
            for placeholder in instr.placeholders  # requested
            if placeholder in ret.placeholders  # available
        ]

        if instr.ids is None:
            records = tuple(tuple(record[i] for i in placeholder_indices) for record in ret.records)
        else:
            ids = set(instr.ids)
            records = tuple(
                tuple(record[i] for i in placeholder_indices)
                for record in ret.records
                if record[ret.id_pos] in ids  # crash on missing IDs
            )
        placeholders = tuple(ret.placeholders[i] for i in placeholder_indices)
        id_pos = placeholders.index(ID) if ID in placeholders else -1
        return PlaceholderTranslations(instr.source, placeholders, records, id_pos)

    def __str__(self) -> str:
        class NoSources:
            def __repr__(self) -> str:
                return "<no sources>"

        sources = self.sources if self.sources else NoSources()
        return_all = self.return_all
        return f"{type(self).__name__}({sources=}, {return_all=})"



import logging
from importlib.metadata import entry_points
from time import perf_counter
from typing import Any

from rics.misc import tname

from ..types import IdType, NameType, SourceType, TranslatableT
from ._data_structure_io import DataStructureIO
from .exceptions import UntranslatableTypeError

AnyDataStructureIO = DataStructureIO[Any, Any, Any, Any]
_RESOLUTION_ORDER: list[type[AnyDataStructureIO]]
LOGGER = logging.getLogger(__package__)


_ENTRYPOINT_GROUP: str = "id_translation.dio"
"""Alias of the ``__init__.ENTRYPOINT_GROUP`` attribute.

Background:
    https://github.com/sphinx-doc/sphinx/issues/6495#issuecomment-1058033697
    https://github.com/sphinx-doc/sphinx/issues/12020
"""


def resolve_io(
    arg: TranslatableT,
    *,
    task_id: int | None = None,
) -> DataStructureIO[TranslatableT, NameType, SourceType, IdType]:
    """Get an IO instance for `arg`.

    Args:
        arg: An argument to get IO for.
        task_id: Used for logging.

    Returns:
        A data structure IO instance for `arg`.

    Raises:
        UntranslatableTypeError: If no suitable IO implementation could be found.

    See Also:
        The :func:`register_io` function.
    """
    eligible_disabled = None
    for io_class in _RESOLUTION_ORDER:
        if io_class.handles_type(arg):
            if io_class.priority < 0 and eligible_disabled is None:
                eligible_disabled = io_class
                continue  # Users may set negative priority to disable implementations after registration.
            return _initialize(arg, io_class, task_id=task_id)

    exc = UntranslatableTypeError(type(arg))
    if eligible_disabled is not None:
        exc.add_note(f"Hint: Implementation '{_pretty_io_name(eligible_disabled)}' is disabled (priority < 0).")

    raise exc


def _initialize(
    arg: TranslatableT,
    io_class: type[DataStructureIO[TranslatableT, NameType, SourceType, IdType]],
    *,
    task_id: int | None = None,
) -> DataStructureIO[TranslatableT, NameType, SourceType, IdType]:
    if LOGGER.isEnabledFor(logging.DEBUG):
        LOGGER.debug(
            f"Using rank-{_RESOLUTION_ORDER.index(io_class)} (priority={io_class.priority}) implementation"
            f" '{_pretty_io_name(io_class)}' for translatable of type='{tname(arg, include_module=True)}'.",
            extra={"task_id": task_id},
        )

    return io_class()


def get_resolution_order(*, real: bool = False) -> list[type[AnyDataStructureIO]]:
    """Returns known :class:`.DataStructureIO` implementations in the correct resolution order.

    Args:
        real: If ``True``, return the actual list instead of a copy.

    Returns:
        A list of IO implementations sorted by rank.
    """
    return _RESOLUTION_ORDER if real else [*_RESOLUTION_ORDER]


def register_io(
    io: type[AnyDataStructureIO],
) -> None:
    """Register a new IO implementation.

    Classes are polled through :meth:`.DataStructureIO.handles_type` in based on :attr:`DataStructureIO.priority`.

    Args:
        io: A :class:`.DataStructureIO` type
    """
    if io.priority < 0:
        LOGGER.warning(f"Refusing to register '{_pretty_io_name(io)}' since priority={io.priority} < 0.")
        return

    _RESOLUTION_ORDER.append(io)

    if LOGGER.isEnabledFor(logging.DEBUG):
        LOGGER.debug(f"Registered IO implementation: '{_pretty_io_name(io)}'.")

    _finalize_order()


def _finalize_order() -> None:
    new_order = set()
    for dio in _RESOLUTION_ORDER:
        if dio.priority < 0:
            continue
        new_order.add(dio)

    _RESOLUTION_ORDER.clear()
    _RESOLUTION_ORDER.extend(new_order)
    _RESOLUTION_ORDER.sort(key=lambda io: io.priority, reverse=True)


def _pretty_io_name(io: type[AnyDataStructureIO]) -> str:
    return tname(io, prefix_classname=True, include_module=True)


def is_registered(io: type[AnyDataStructureIO]) -> bool:
    """Return IO implementation registration status.

    Implementations should register themselves using :meth:`.DataStructureIO.register`.

    Args:
        io: A :class:`.DataStructureIO` type.
    """
    return io.priority >= 0 and io in _RESOLUTION_ORDER


def load_integrations() -> None:
    """Discover, load and register entrypoint integrations.

    Reset the registry, then load entrypoints in the
    :const:`{_ENTRYPOINT_GROUP!r} <id_translation.dio.ENTRYPOINT_GROUP>`
    entrypoint group (see :py:func:`importlib.metadata.entry_points` for details).

    Will skip integrations that raise :class:`ImportError` when loaded.

    Raises:
        TypeError: If an integration does not inherit from :class:`.DataStructureIO`.

    Notes:
        Called automatically when :mod:`id_translation` is imported.
    """
    from .default import DictIO, SequenceIO, SetIO, SingleValueIO  # noqa: PLC0415

    global _RESOLUTION_ORDER  # noqa: PLW0603

    _RESOLUTION_ORDER = [DictIO, SetIO, SequenceIO, SingleValueIO]

    start = perf_counter()

    LOGGER.debug("Initializing %s integrations in group='%s'.", DataStructureIO.__name__, _ENTRYPOINT_GROUP)

    n_total = 0
    n_ok = 0
    for ep in entry_points(group=_ENTRYPOINT_GROUP):
        n_total += 1

        try:
            cls = ep.load()
        except ImportError as e:
            if LOGGER.isEnabledFor(logging.DEBUG):
                LOGGER.debug(f"Failed to import entrypoint={ep!r}: {e!r}.")
            continue

        if not issubclass(cls, DataStructureIO):
            msg = f"Bad entrypoint={ep!r}: {cls} is not a subtype of {DataStructureIO.__name__}. "
            raise TypeError(msg)

        _RESOLUTION_ORDER.append(cls)
        n_ok += 1

    _finalize_order()

    millis = round(1000 * (perf_counter() - start))
    LOGGER.debug("Imported and registered %i/%i integrations in %i milliseconds.", n_ok, n_total, millis)


if load_integrations.__doc__:
    load_integrations.__doc__ = load_integrations.__doc__.format(_ENTRYPOINT_GROUP=_ENTRYPOINT_GROUP)



from collections.abc import Sequence
from typing import Any

from id_translation.offline import TranslationMap
from id_translation.types import IdType, NameType, SourceType

from .._data_structure_io import DataStructureIO


class SetIO(DataStructureIO[set[IdType], NameType, SourceType, IdType]):
    """IO implementation for ``set`` types."""

    priority = 1010

    @classmethod
    def handles_type(cls, arg: Any) -> bool:
        return isinstance(arg, set)

    @classmethod
    def extract(cls, translatable: set[IdType], names: list[NameType]) -> dict[NameType, Sequence[IdType]]:
        if len(names) != 1:  # pragma: no cover
            raise ValueError("Length of names must be one.")

        return {names[0]: list(translatable)}

    @classmethod
    def insert(
        cls,
        translatable: set[IdType],
        names: list[NameType],
        tmap: TranslationMap[NameType, SourceType, IdType],
        copy: bool,
    ) -> set[str] | None:
        magic_dict = tmap[names[0]]
        translated = {magic_dict[e] for e in translatable}

        if copy:
            return translated

        translatable.clear()
        translatable.update(translated)  # type: ignore[arg-type]
        return None



from collections.abc import Iterator, Mapping, Sequence
from copy import copy
from typing import TYPE_CHECKING, Any, Generic, Self

from rics.collections.dicts import InheritedKeysDict, reverse_dict
from rics.misc import tname

from ..transform.types import Transformer, Transformers
from ..types import HasSources, IdType, NameToSource, NameType, SourceType
from ._format import Format
from ._format_applier import FormatApplier
from ._magic_dict import MagicDict
from .types import FormatType, PlaceholderTranslations, SourcePlaceholderTranslations

if TYPE_CHECKING:
    import pandas


class TranslationMap(
    Generic[NameType, SourceType, IdType],
    HasSources[SourceType],
    Mapping[NameType | SourceType, MagicDict[IdType]],
):
    """Storage class for fetched translations.

    Args:
        source_translations: Fetched translations ``{source: PlaceholderTranslations}``.
        name_to_source: Mappings ``{name: source}``, but may be overridden by the user.
        fmt: A translation format. Must be given to use as a mapping.
        default_fmt: Alternative format specification to use instead of `fmt` for fallback translation.
        default_fmt_placeholders: Per-source default placeholder values.
        enable_uuid_heuristics: Improves matching when :py:class:`~uuid.UUID`-like IDs are in use.
        transformers: A dict ``{source: transformer}`` of initialized :class:`.Transformer` instances.

    """

    def __init__(
        self,
        source_translations: SourcePlaceholderTranslations[SourceType],
        *,
        fmt: FormatType = Format.DEFAULT,
        default_fmt: FormatType = Format.DEFAULT_FAILED,
        name_to_source: NameToSource[NameType, SourceType] | None = None,
        default_fmt_placeholders: InheritedKeysDict[SourceType, str, Any] | None = None,
        enable_uuid_heuristics: bool = True,
        transformers: Transformers[SourceType, IdType] | None = None,
    ) -> None:
        self.fmt = Format.parse(fmt)
        self.default_fmt = Format.parse(default_fmt)
        self.default_fmt_placeholders = default_fmt_placeholders or InheritedKeysDict.make({})

        transformers = transformers or {}
        self._format_appliers: dict[SourceType, FormatApplier[NameType, SourceType, IdType]] = {
            source: FormatApplier(translations, transformer=transformers.get(source))
            for source, translations in source_translations.items()
        }
        self._names_and_sources: set[NameType | SourceType] = set()
        self.name_to_source = name_to_source or {}

        self._reverse_mode: bool = False
        self.enable_uuid_heuristics = enable_uuid_heuristics

    def to_dicts(self) -> dict[SourceType, dict[str, Sequence[Any]]]:
        """Get the underlying data used for translations as dicts.

        This is equivalent using :meth:`to_pandas`, then calling ``DataFrame.to_dict(orient='list')`` on each frame.

        Returns:
            A dict ``{source: {placeholder: [values...]}}``.
        """
        return {applier.source: applier.to_dict() for applier in self._format_appliers.values()}

    def to_pandas(self) -> dict[SourceType, "pandas.DataFrame"]:
        """Get the underlying data used for translations as :class:`pandas.DataFrame`.

        Returns:
            A dict ``{source: DataFrame}``.
        """
        return {applier.source: applier.to_pandas() for applier in self._format_appliers.values()}

    def to_translations(self, fmt: FormatType | None = None) -> dict[SourceType, MagicDict[IdType]]:
        """Create translations for all sources.

        Returned values are of type :class:`.MagicDict`. To convert to regular built-in dicts, run

        .. code-block::

           translations = translation_map.to_translations()
           as_regular_dicts = {
              source: dict(magic)
              for source, magic in translations.items()
           }

        on the returned dict-of-magic-dicts.

        Args:
            fmt: :class:`.Format` to use. If ``None``, fall back to init format.

        Returns:
            A dict of translations ``{source: MagicDict}``.
        """
        return {source: self.apply(source, fmt=fmt) for source in self.sources}

    @classmethod
    def from_pandas(
        cls,
        frames: dict[SourceType, "pandas.DataFrame"],
        fmt: FormatType = Format.DEFAULT,
        *,
        default_fmt: FormatType = Format.DEFAULT_FAILED,
    ) -> Self:
        """Create a new instance from a :class:`pandas.DataFrame` dict.

        Args:
            frames: A dict ``{source: DataFrame}``.
            fmt: A translation format. Must be given to use as a mapping.
            default_fmt: Alternative format specification to use instead of `fmt` for fallback translation.

        Returns:
            A new ``TranslationMap``.
        """
        source_translations = {
            source: PlaceholderTranslations.from_dataframe(source, frame) for source, frame in frames.items()
        }
        return cls(source_translations, fmt=fmt, default_fmt=default_fmt)

    def apply(
        self,
        name_or_source: NameType | SourceType,
        fmt: FormatType | None = None,
        *,
        default_fmt: FormatType | None = None,
    ) -> MagicDict[IdType]:
        """Create translations for a given name or source.

        Args:
            name_or_source: A name or source to translate.
            fmt: :class:`.Format` to use. If ``None``, fall back to init format.
            default_fmt: Alternative format for default translation. Resolution: Arg -> init arg, fmt arg, init fmt arg

        Returns:
            Translations for `name` as a dict ``{id: translation}``.

        Raises:
            ValueError: If ``fmt=None`` and initialized without `fmt`.
            KeyError: If trying to translate `name` which is not known.

        Notes:
             This method is called by ``__getitem__``.
        """
        fmt = self._fmt if fmt is None else fmt
        if fmt is None:
            raise ValueError("No format specified and None given at initialization.")  # pragma: no cover

        fmt = Format.parse(fmt)
        default_fmt = self._default_fmt if default_fmt is None else Format.parse(default_fmt)
        source = self.name_to_source.get(name_or_source, name_or_source)  # type: ignore
        translations: MagicDict[IdType] = self._format_appliers[source](
            fmt,
            default_fmt=default_fmt,
            default_fmt_placeholders=self.default_fmt_placeholders.get(source),
            enable_uuid_heuristics=self.enable_uuid_heuristics,
        )

        return (
            MagicDict(
                reverse_dict(translations, duplicate_key_action="raise"),  # type: ignore
                default_value=default_fmt.fstring(positional=True),
                enable_uuid_heuristics=False,
            )
            if self.reverse_mode
            else translations
        )

    @property
    def names(self) -> list[NameType]:
        """Return names that can be translated."""
        return list(self.name_to_source)

    @property
    def sources(self) -> list[SourceType]:
        return list(self._format_appliers)

    @property
    def placeholders(self) -> dict[SourceType, list[str]]:
        return {applier.source: applier.placeholders for applier in self._format_appliers.values()}

    @property
    def len_per_source(self) -> dict[SourceType, int]:
        """Number of IDs per :attr:`source <sources>`."""
        return {applier.source: len(applier) for applier in self._format_appliers.values()}

    @property
    def name_to_source(self) -> NameToSource[NameType, SourceType]:
        """Return name-to-source mapping."""
        return dict(self._name_to_source)

    @name_to_source.setter
    def name_to_source(self, value: NameToSource[NameType, SourceType]) -> None:
        self._name_to_source: NameToSource[NameType, SourceType] = value
        self._names_and_sources = set(value).union(self._format_appliers)

    @property
    def fmt(self) -> Format:
        """Return the translation format."""
        return self._fmt

    @fmt.setter
    def fmt(self, value: FormatType) -> None:
        self._fmt = Format.parse(value)

    @property
    def default_fmt(self) -> Format:
        """Return the format specification to use instead of :attr:`fmt` for fallback translation."""
        return self._default_fmt

    @default_fmt.setter
    def default_fmt(self, value: FormatType) -> None:
        self._default_fmt = Format.parse(value)

    @property
    def default_fmt_placeholders(self) -> InheritedKeysDict[SourceType, str, Any]:
        """Return the default translations used for `default_fmt_placeholders` placeholders."""
        return self._default_fmt_placeholders

    @default_fmt_placeholders.setter
    def default_fmt_placeholders(self, value: InheritedKeysDict[SourceType, str, Any]) -> None:
        self._default_fmt_placeholders = InheritedKeysDict.make(value) if value else InheritedKeysDict()

    @property
    def reverse_mode(self) -> bool:
        """Return reversed mode status flag.

         If set, the mappings returned by :meth:`apply` (and therefore also ``__getitem__``) are reversed.

        Returns:
            Reversal status flag.
        """
        return self._reverse_mode

    @reverse_mode.setter
    def reverse_mode(self, value: bool) -> None:
        self._reverse_mode = value

    @property
    def enable_uuid_heuristics(self) -> bool:
        """Return automatic UUID mitigation status."""
        return self._enable_uuid_heuristics

    @enable_uuid_heuristics.setter
    def enable_uuid_heuristics(self, value: bool) -> None:
        self._enable_uuid_heuristics = value

    @property
    def transformers(self) -> dict[SourceType, Transformer[IdType]]:
        """Get a dict ``{source: transformer}`` of :class:`.Transformer` instances used by this ``TranslationMap``."""
        return {
            source: applier.transformer
            for source, applier in self._format_appliers.items()
            if applier.transformer is not None
        }

    def copy(self) -> Self:
        """Make a copy of this ``TranslationMap``."""
        return copy(self)

    def __getitem__(self, name_or_source: NameType | SourceType) -> MagicDict[IdType]:
        return self.apply(name_or_source)

    def __len__(self) -> int:
        return len(self._names_and_sources)

    def __iter__(self) -> Iterator[NameType | SourceType]:
        return iter(self._names_and_sources)

    def __bool__(self) -> bool:
        return bool(self._format_appliers)

    def __repr__(self) -> str:
        len_per_source = self.len_per_source
        order = sorted(len_per_source, key=len_per_source.__getitem__, reverse=True)
        sources = ", ".join(f"'{source}': {len_per_source[source]} IDs" for source in order)
        return f"{tname(self)}({sources})"



"""Utility module for parsing raw ``Format`` input strings."""

import collections.abc as _abc
from dataclasses import dataclass as _dataclass
from string import Formatter as _Formatter
from typing import Any as _Any
from typing import NamedTuple as _NamedTuple

OPTIONAL_BLOCK_START_DELIMITER = "["
_START = OPTIONAL_BLOCK_START_DELIMITER
OPTIONAL_BLOCK_END_DELIMITER = "]"
_END = OPTIONAL_BLOCK_END_DELIMITER


class ParseBlockResult(_NamedTuple):
    """Output type of :meth:`.Element.parse_block`."""

    parsed_block: str
    """Processed parts of the block."""
    placeholders: list[str]
    """Names of the :attr:`.Format.placeholders` in `parsed_block`, in the order in which they appear."""


class MalformedOptionalBlockError(ValueError):
    """Error raised for improper optional blocks."""

    @staticmethod
    def get_marker_row(format_string: str, open_idx: int = -1, idx: int = -1) -> str:
        """Get a string of length equal to `format_string` which marks problem locations."""
        markers = [" "] * len(format_string)

        if idx != -1:
            markers[idx] = "^"
        if open_idx != -1:
            markers[open_idx] = "^"
        return "".join(markers)


class BadDelimiterError(MalformedOptionalBlockError):
    """Errors raised due to mismatched delimiters."""

    def __init__(self, format_string: str, open_idx: int, idx: int) -> None:
        problem_locations = MalformedOptionalBlockError.get_marker_row(format_string, open_idx, idx)

        if idx == -1:
            super().__init__(
                f"""Malformed optional block. Optional block opened at i={open_idx} was never closed.
'{format_string}'
 {problem_locations}""".strip()
            )
        else:
            hint = (
                f"Hint: Use double characters to escape '{_START}' and '{_END}', "
                f"e.g. '{_START * 2}' to render a single '{_START}'-character."
            )
            info = (
                "there is no block to close"
                if open_idx == -1
                else f"nested optional blocks are not supported (opened at {open_idx})"
            )

            super().__init__(
                f"""Malformed optional block. Got '{format_string[idx]}' at i={idx}, but {info}.
    '{format_string}'
     {problem_locations}
{hint}""".strip()
            )


@_dataclass(frozen=True)
class Element:
    """Information about a single block in a ``Format`` specification."""

    part: str
    """String literal."""
    placeholders: list[str]
    """Placeholder names in `part`, if any."""
    required: bool
    """Flag indicating whether the element may be excluded."""
    positional_part: str
    """Return a positional version of the `part` attribute."""

    @classmethod
    def make(cls, fmt: str, in_optional_block: bool) -> "Element":
        """Create an ``Element`` from an input string `s`.

        Args:
            fmt: Input data.
            in_optional_block: Flag indicating whether `s` was found inside an optional block.

        Returns:
            A new ``Element``.
        """
        block = fmt
        fmt = block.replace("[[", "[").replace("]]", "]")

        positional_part, placeholders = cls.parse_block(fmt)
        is_optional = placeholders and in_optional_block

        return Element(fmt, placeholders, required=not is_optional, positional_part=positional_part)

    @classmethod
    def parse_block(cls, block: str, defaults: _abc.Mapping[str, _Any] | None = None) -> ParseBlockResult:
        """Parse an entire block with optional defaults for placeholders found.

        .. hint::

           With `defaults`, the value of the :attr:`ParseBlockResult.parsed_block` is more or less what you'd expect if
           the built-in :py:meth:`str.format_map`-method allowed missing keys.

        Anonymous fields are not permitted.

        ..
           >>> from uuid import UUID

        Output with ``defaults == None``:
            When `defaults` are ``None``, placeholder names are stripped from the
            :attr:`~.ParseBlockResult.parsed_block`.

            >>> block, placeholders = Element.parse_block("{id!s:.8}:{name!r}")
            >>> print(f"{block=} has {placeholders=}")
            block='{!s:.8}:{!r}' has placeholders=['id', 'name']

            Field names in `block` are returned as :attr:`.ParseBlockResult.placeholders`, in the order in which they
            appeared in the input `block`. The field names of :attr:`~.ParseBlockResult.parsed_block` will be
            anonymous.

            >>> block.format(UUID(int=10**38), "Morran Borran")
            "4b3b4ca8:'Morran Borran'"

        Output with ``defaults != None``:
            When `defaults` are given, all placeholders in the `block` which are present in the `defaults` are replaced
            with ``defaults[field_name]`` in the :attr:`~.ParseBlockResult.parsed_block`.

            >>> block, placeholders = Element.parse_block(
            ...     "{id!s:.8}:{name!r}",
            ...     defaults={"name": "Morran Borran"},
            ... )
            >>> print(f"{block=} has {placeholders=}")
            block="{id!s:.8}:'Morran Borran'" has placeholders=['id']

            Field names without defaults will be present both in :attr:`~.ParseBlockResult.placeholders`, and as named
            fields in the :attr:`~.ParseBlockResult.parsed_block`.

            >>> block.format(id=UUID(int=10**38))
            "4b3b4ca8:'Morran Borran'"

        Args:
            block: A block to parse.
            defaults: A dict ``{placeholder: value}``.

        Returns:
            A :class:`.ParseBlockResult` tuple.

        Raises:
            ValueError: If `block` contains anonymous fields.
        """
        if defaults is None:
            defaults = {}
            keep_placeholder_names = False
        else:
            keep_placeholder_names = True

        parts = []
        placeholders = []

        # Optional block backtracking
        replaced_parts_index = []
        placeholders_index = []

        formatter = _Formatter()

        for literal_text, field_name, format_spec, conversion in formatter.parse(block):
            parts.append(literal_text.replace("{", "{{").replace("}", "}}"))

            if field_name == "":
                msg = f"Bad {block=}; anonymous fields are not permitted."
                msg += "\n- Hint: Replace '{}' with '{field_name}' to give this field a name"
                msg += "\n- Hint: Replace '{}' with '{{}}' to render literal curly braces"
                raise ValueError(msg)

            if field_name:
                placeholder, _, attribute = field_name.partition(".")
                formatting_parts = _get_formatting_parts(
                    attribute=attribute, conversion=conversion, format_spec=format_spec
                )

                if placeholder in defaults:
                    replaced_parts_index.append(len(parts))
                    formatting = "".join(["{", *formatting_parts, "}"])
                    value = defaults[placeholder]
                    parts.append(formatting.format(value))
                else:
                    parts.append("{")

                    placeholders.append(placeholder)
                    if keep_placeholder_names:
                        placeholders_index.append(len(parts))
                        parts.append(placeholder)

                    parts.extend(formatting_parts)
                    parts.append("}")

        return ParseBlockResult("".join(parts), placeholders=placeholders)


def _get_delimiter_index(part: str, *, start: bool) -> int | None:
    def _exec(func: _abc.Callable[[str], int], sign: str) -> int | None:
        try:
            return func(sign)
        except ValueError:
            return None

    if start:
        return _exec(part.index, _START)
    else:
        return _exec(part.rindex, _END)


def _get_formatting_parts(
    *,
    attribute: str | None,
    conversion: str | None,
    format_spec: str | None,
) -> _abc.Iterable[str]:
    if attribute:
        yield "."
        yield attribute
    if conversion is not None:
        yield "!"
        yield conversion
    if format_spec is not None and format_spec != "":
        yield ":"
        yield format_spec


def get_elements(fmt: str) -> list[Element]:  # noqa: PLR0912
    """Split a format string into elements.

    Args:
        fmt: User input string.

    Returns:
        A list of parsed elements.

    Raises:
        BadDelimiterError: For unbalanced optional block delimitation characters.
    """
    if not fmt:
        return [Element("", [], required=True, positional_part="")]

    same_count = 1
    ans = []

    in_optional_block = fmt[0] == _START
    open_idx = 0 if in_optional_block else -1
    prev_idx = int(in_optional_block)

    for idx in range(int(in_optional_block), len(fmt)):
        char = fmt[idx]
        next_char = fmt[idx + 1] if idx + 1 < len(fmt) else None
        is_delimiter_char = char in (_START, _END)
        if next_char == char and is_delimiter_char:
            same_count += 1
        else:
            if same_count % 2 and is_delimiter_char:
                if char == _START:
                    if open_idx != -1:
                        raise BadDelimiterError(fmt, open_idx, idx)
                    open_idx = idx
                else:
                    if open_idx == -1:
                        raise BadDelimiterError(fmt, open_idx, idx)
                    open_idx = -1

                if prev_idx != idx:
                    part = fmt[prev_idx:idx]
                    element = Element.make(part, in_optional_block)
                    if in_optional_block and not element.placeholders:
                        # Convert to required
                        element = Element.make(_START + part + _END, True)
                        assert element.required is True  # noqa: S101
                        assert len(element.placeholders) == 0  # noqa: S101
                        assert element.part == element.positional_part  # noqa: S101
                    ans.append(element)
                in_optional_block = not in_optional_block
                prev_idx = idx + 1

            same_count = 1

    if prev_idx != len(fmt):
        ans.append(Element.make(fmt[prev_idx:], in_optional_block))

    if in_optional_block:
        raise BadDelimiterError(fmt, open_idx, -1)

    return ans



"""Backend for the :meth:`.Translator.from_config` method."""

from . import factories, meta
from ._factory import TranslatorFactory
from ._load_toml import load_toml_file

__all__ = [
    "TranslatorFactory",
    "factories",
    "load_toml_file",
    "meta",
]



from typing import Any

from id_translation.fetching import CacheAccess
from id_translation.types import IdType, SourceType

from ._initialize import initialize


def default_cache_access_factory(clazz: str, config: dict[str, Any]) -> CacheAccess[SourceType, IdType]:
    """Create a :class:`.CacheAccess` from config."""
    return initialize(
        clazz,
        config,
        CacheAccess,  # type: ignore[type-abstract]  # https://github.com/python/mypy/issues/4717
        default_module=None,
    )



"""Types that are specific to the :class:`.Translator` implementation."""

import typing as _t
from collections import abc as _abc

from rics.collections.dicts import MakeType as _MakeType

from . import types as _tt
from .fetching import CacheAccess as _CacheAccess
from .fetching import Fetcher as _Fetcher
from .mapping import Mapper as _Mapper
from .mapping.types import UserOverrideFunction as _UserOverrideFunction
from .offline import TranslationMap as _TranslationMap
from .offline import types as _ot
from .transform.types import Transformer as _Transformer

SimpleDictFetcherTypes = dict[_tt.SourceType, dict[_tt.IdType, str]]
"""Data for translating using only the default `id` and `name` placeholders.

Must be on the form ``{source: {id: name}}``.
"""

SourceToPlaceholderTranslationsMakeTypes = _abc.Mapping[_tt.SourceType, _ot.MakeTypes[_tt.SourceType, _tt.IdType]]
"""Data for translating using arbitrary placeholders; see :meth:`.PlaceholderTranslations.make`"""

NativeFetcherTypes: _t.TypeAlias = (
    _TranslationMap[_tt.NameType, _tt.SourceType, _tt.IdType]
    | _Fetcher[_tt.SourceType, _tt.IdType]
    | _ot.SourcePlaceholderTranslations[_tt.SourceType]
)
"""Internal types related to fetching."""

FetcherTypes: _t.TypeAlias = (
    NativeFetcherTypes[_tt.NameType, _tt.SourceType, _tt.IdType]
    | SimpleDictFetcherTypes[_tt.SourceType, _tt.IdType]
    | SourceToPlaceholderTranslationsMakeTypes[_tt.SourceType, _tt.IdType]
)
"""All valid input types for creating a ``Translator``."""


class AbstractFetcherParams(_t.TypedDict, _t.Generic[_tt.SourceType, _tt.IdType], total=False):
    """Keyword arguments for the :class:`.AbstractFetcher` base class."""

    mapper: _Mapper[str, str, _tt.SourceType] | None
    allow_fetch_all: bool
    selective_fetch_all: bool
    identifiers: _abc.Sequence[str] | None
    optional: bool
    cache_access: _CacheAccess[_tt.SourceType, _tt.IdType] | None


class MapParams(_t.TypedDict, _t.Generic[_tt.NameType, _tt.SourceType, _tt.IdType], total=False):
    """Arguments of :meth:`.Translator.map` and :meth:`.Translator.map_scores`."""

    translatable: _t.Required[_tt.Translatable[_tt.NameType, _tt.IdType]]
    names: _tt.NameTypes[_tt.NameType] | None
    ignore_names: _tt.Names[_tt.NameType] | None
    override_function: _UserOverrideFunction[_tt.NameType, _tt.SourceType, None] | None


class UniqueCopyParams(_t.TypedDict, _t.Generic[_tt.NameType, _tt.SourceType, _tt.IdType], total=False):
    """Arguments of :meth:`.Translator.copy` that do not overlap with :class:`AllTranslateParams`."""

    fetcher: FetcherTypes[_tt.NameType, _tt.SourceType, _tt.IdType]
    mapper: _Mapper[_tt.NameType, _tt.SourceType, None]
    default_fmt: _ot.FormatType
    default_fmt_placeholders: _MakeType[_tt.SourceType, str, _t.Any] | None
    enable_uuid_heuristics: bool
    transformers: dict[_tt.SourceType, _Transformer[_tt.IdType]] | None


class CopyParams(UniqueCopyParams[_tt.NameType, _tt.SourceType, _tt.IdType], total=False):
    """All arguments of :meth:`.Translator.copy`."""

    fmt: _ot.FormatType


class TranslateParams(_t.TypedDict, _t.Generic[_tt.NameType, _tt.SourceType, _tt.IdType], total=False):
    """Keyword arguments of :meth:`.Translator.translate`.

    .. note::

       Does not include `translatable` or `copy`.


    **Motivation**

    Allowing these to be passed as keyword arguments causes issues with typing, especially method overloading. For
    example:

    .. code-block::

       def func(**kwargs: Unpack[AllTranslateParams]):
          translatable = kwargs["translatable"]
          if isinstance(translatable, list) and kwargs.get(copy, True):
              raise CustomException("we don't do that here")

          else:
              # Do whatever

    using :class:`AllTranslateParams` is typically not as safe as:

    .. code-block::

       @overload
       def func(...): ...

       @overload
       def func(
           translatable: list, copy: Literal[False],
           **kwargs: Unpack[TranslateParams]
       ) -> Never: ...

       def func(translatable, copy, **kwargs: Unpack[TranslateParams]):
           "Implementation as above"

    since ``func(translatable=[], copy=False)`` does not behave like
    ``Translator.translate([], copy=False)`` would. Functions that transparently wrap
    :meth:`.Translator.translate` should probably use :py:func:`functools.wraps` instead.
    """

    names: _tt.NameTypes[_tt.NameType] | _tt.NameToSource[_tt.NameType, _tt.SourceType] | None
    ignore_names: _tt.Names[_tt.NameType] | None
    override_function: _UserOverrideFunction[_tt.NameType, _tt.SourceType, None] | None
    max_fails: float
    reverse: bool
    fmt: _ot.FormatType | None


class AllTranslateParams(TranslateParams[_tt.NameType, _tt.SourceType, _tt.IdType], total=False):
    """All arguments of :meth:`.Translator.translate`."""

    translatable: _t.Required[_tt.Translatable[_tt.NameType, _tt.IdType]]
    copy: bool


class FetchParams(_t.TypedDict, _t.Generic[_tt.NameType, _tt.SourceType, _tt.IdType], total=False):
    """All arguments of :meth:`.Translator.fetch` and :meth:`.Translator.go_offline`."""

    translatable: _tt.Translatable[_tt.NameType, _tt.IdType] | None
    names: _tt.NameTypes[_tt.NameType] | _tt.NameToSource[_tt.NameType, _tt.SourceType] | None
    ignore_names: _tt.Names[_tt.NameType] | None
    override_function: _UserOverrideFunction[_tt.NameType, _tt.SourceType, None] | None
    fmt: _ot.FormatType | None


CacheMissReasonType = _t.Literal["metadata-missing", "metadata-changed", "too-old"]
"""High-level reason why a cached ``Translator`` was rejected."""



from collections.abc import Hashable, Iterable
from math import isfinite
from typing import TYPE_CHECKING, Any, Generic, TypeVar

from ..types import CandidateType, ValueType

T = TypeVar("T", bound=Hashable)


if TYPE_CHECKING:
    import pandas

inf = float("inf")


class ScoreMatrix(Generic[ValueType, CandidateType]):
    """A matrix of match scores.

    Args:
        values: Iterable of elements to match to candidates.
        candidates: Iterable of candidates to match with `value`. Duplicate elements will be discarded.
        grid: Initial score matrix. Default is to fill with ``-inf``.

    Raises:
        ValueError: If a bad `grid` is given.
    """

    def __init__(
        self,
        values: Iterable[ValueType],
        candidates: Iterable[CandidateType],
        *,
        grid: list[list[float]] | None = None,
    ) -> None:
        self._values = _deduplicate(values)
        self._candidates = _deduplicate(candidates)

        if grid is None:
            grid = self._new_grid(-inf)
        else:
            _verify_user_grid(grid, self._values, len(self._candidates))

        self._grid = grid

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self._values!r}, {self._candidates!r}, grid={self._grid})"

    def _new_grid(self, value: float) -> list[list[float]]:
        row = [value] * len(self._candidates)
        return [[*row] for _ in range(len(self._values))]

    def __setitem__(self, index: tuple[ValueType | slice, CandidateType | slice], value: float) -> None:
        i, j = index

        if __debug__:
            if isinstance(i, slice) and i != slice(None, None, None):
                msg = f"slice {index=} not supported"
                raise NotImplementedError(msg)
            if isinstance(j, slice) and j != slice(None, None, None):
                msg = f"slice {index=} not supported"
                raise NotImplementedError(msg)

        if isinstance(i, slice):
            if isinstance(j, slice):
                # All rows and columns.
                self._grid = self._new_grid(value)
            else:
                gj = self._candidate_index(j)
                for row in self._grid:
                    row[gj] = value
        else:
            gi = self._value_index(i)
            if isinstance(j, slice):
                self._grid[gi] = [value for _ in range(len(self._candidates))]
            else:
                gj = self._candidate_index(j)
                self._grid[gi][gj] = value

    def _value_index(self, value: ValueType) -> int:
        try:
            return self._values.index(value)
        except ValueError:
            pass

        # Add missing element.
        self._values.append(value)
        self._grid.append([-inf for _ in range(len(self._candidates))])
        return self._values.index(value)

    def _candidate_index(self, candidate: CandidateType) -> int:
        try:
            return self._candidates.index(candidate)
        except ValueError:
            pass

        # Add missing element.
        self._candidates.append(candidate)
        for row in self._grid:
            row.append(-inf)
        return self._candidates.index(candidate)

    @property
    def size(self) -> int:
        """Total number of elements."""
        return len(self._values) * len(self._candidates)

    @property
    def values(self) -> list[ValueType]:
        """Unique values in order."""
        return [*self._values]

    @property
    def candidates(self) -> list[CandidateType]:
        """Unique candidates in order."""
        return [*self._candidates]

    def get_finite_values(self) -> set[ValueType]:
        """Compute all finite values."""
        return {value for value, row in zip(self._values, self._grid, strict=True) if all(map(isfinite, row))}

    def to_pandas(self) -> "pandas.DataFrame":
        """Convert to :class:`pandas.DataFrame`."""
        from pandas import DataFrame, Index  # noqa: PLC0415

        return DataFrame(
            self._grid,
            index=Index(self._values, name="values"),
            columns=Index(self._candidates, name="candidates"),
        )

    def to_dict(self) -> dict[tuple[ValueType, CandidateType], float]:
        """Convert to dict ``{(value, candidate): score}``."""
        rv = {}
        for value, row in zip(self._values, self._grid, strict=True):
            for candidate, score in zip(self._candidates, row, strict=True):
                rv[(value, candidate)] = score

        return rv

    def to_string(self, *, decimals: int = 2) -> str:
        """Format score table."""
        try:
            return self.to_pandas().to_string(float_format=f"%.{decimals}f")  # type: ignore[no-any-return]
        except ImportError:
            return self.to_native_string(decimals=decimals)

    def to_native_string(self, *, decimals: int = 2, lines: bool = True) -> str:
        """Format score table without ``pandas``."""
        column_separator = " ┃ " if lines else " "

        header_row = ["v/c"] + [str(c) for c in self._candidates]
        formatted_scores = [[f"{float(s):.{decimals}f}" for s in scores] for scores in self._grid]
        width = max(*map(len, header_row + [str(i) for i in self._values]), *map(len, formatted_scores))

        rows = []
        for value, scores in zip(self._values, formatted_scores, strict=False):
            row = [f"{value:<{width}}"] + [f"{s:>{width}}" for s in scores]
            rows.append(column_separator.join(row))

        column_headers = [f"{h:<{width}}" for h in header_row]
        header = [column_separator.join(column_headers)]
        if lines:
            horizontal_line = ["━" * len(h) for h in column_headers]
            header.append("━╋━".join(horizontal_line))

        return "\n".join(header + rows)


def _deduplicate(items: Iterable[T]) -> list[T]:
    seen = set()
    unique = []
    for item in items:
        if item in seen:
            continue

        seen.add(item)
        unique.append(item)

    return unique


def _verify_user_grid(grid: list[list[float]], values: list[Any], candidates: int) -> None:
    nvalues = len(values)
    if len(grid) != nvalues:
        msg = f"Bad grid: Number of rows {len(grid)} must match the number of values={nvalues}."
        raise ValueError(msg)

    for i, (value, row) in enumerate(zip(values, grid, strict=True)):
        if len(row) != candidates:
            msg = f"Bad grid[{i}] row ({value=}): Number of columns {len(row)} must match the number of {candidates=}."
            raise ValueError(msg)



"""Integration for `Dask <https://www.dask.org/>`_ types."""

import typing as _t
from collections import abc as _abc

from dask import compute as _compute
from dask import dataframe as _dd

from id_translation import dio as _dio
from id_translation import types as _tt
from id_translation.dio.exceptions import NotInplaceTranslatableError as _NotInplaceTranslatableError
from id_translation.offline import MagicDict as _MagicDict
from id_translation.offline import TranslationMap as _TranslationMap

DaskT = _t.TypeVar("DaskT", _dd.DataFrame, _dd.Series)
"""Supported ``dask`` types."""


class DaskIO(_dio.DataStructureIO[DaskT, str, _tt.SourceType, _tt.IdType]):
    """Optional IO implementation for ``dask`` types."""

    priority = 1980

    @classmethod
    def handles_type(cls, arg: _t.Any) -> bool:
        return isinstance(arg, (_dd.DataFrame, _dd.Series))

    @classmethod
    def names(cls, translatable: DaskT) -> list[str] | None:
        if isinstance(translatable, _dd.DataFrame):
            return list(translatable.columns)

        return None if translatable.name is None else [translatable.name]

    @classmethod
    def extract(cls, translatable: DaskT, names: list[str]) -> dict[str, _abc.Sequence[_tt.IdType]]:
        if isinstance(translatable, _dd.Series):
            if len(names) != 1:
                raise RuntimeError(f"{len(names)=} != 1 is not supported for dask.Series")
            return {names[0]: translatable.unique().compute().to_list()}
        else:
            delayed = {n: translatable[n].unique() for n in names}
            name_to_ids = _compute(delayed)[0]
            return {name: ids.to_list() for name, ids in name_to_ids.items()}

    @classmethod
    def insert(
        cls,
        translatable: DaskT,
        names: list[str],
        tmap: _TranslationMap[str, _tt.SourceType, _tt.IdType],
        copy: bool,
    ) -> DaskT | None:
        if not copy:
            raise _NotInplaceTranslatableError(translatable)  # Can't in-place a compute graph.

        if isinstance(translatable, _dd.Series):
            return _translate_series(translatable, tmap[names[0]])
        else:
            return _translate_frame(translatable, names, tmap)


def _translate_series(series: _dd.Series, magic_dict: _MagicDict[_tt.IdType]) -> _dd.Series:
    mapping = {idx: magic_dict[idx] for idx in series.unique().compute()}
    return series.replace(mapping)  # type: ignore[no-any-return]


def _translate_frame(
    df: _dd.DataFrame,
    names: list[_tt.NameType],
    tmap: _TranslationMap[_tt.NameType, _tt.SourceType, _tt.IdType],
) -> _dd.DataFrame:
    original_columns = df.columns

    try:
        df.columns = [str(i) for i in range(len(original_columns))]
        for name, int_col in zip(original_columns, df.columns, strict=True):
            if name in names:
                translated = _translate_series(df[int_col], tmap[name])
                df[int_col] = translated
    finally:
        df.columns = original_columns

    return df



"""Transformations for translating bitmask fields."""

import typing as _t
from collections import abc as _abc

from rics.misc import format_kwargs as _format_kwargs

from ..types import Transformer as _Transformer

IdType = int


class TomlOverrideRecord(_t.TypedDict):
    id: IdType
    override: str


class BitmaskTransformer(_Transformer[IdType]):
    r"""Transformations for translating bitmask fields.

    IDs must be integers.

    .. important::

       When using TOML config, dict keys must be strings. Use alternative format for `overrides`:

       .. code-block:: toml

          [transform.'<source>'.BitmaskTransformer]
          overrides = [
            { id = 0, override = "override-for-id=0" },
            { id = 1, override = "override-for-id=1" },
          ]

       Key names must match exactly, and IDs may not be repeated. For more information about TOML configuration, see
       :ref:`translator-config-transform`.

    Args:
        joiner: A string used to join bitmask flag labels.
        overrides: A dict ``{id: translation}``. Use to add or override the translation source.
        force_decomposition: If ``True``, ignore composite values in the translation source.
        force_real_translations: If ``True``, convert :class:`.MagicDict` instances to plain ``dict`` using the
            :attr:`.MagicDict.real` attribute. Results such as ``'<Failed: id=2> & 4:name-of-4'`` are possible when
            ``False``, and will be considered hits by :meth:`translate(max_fails \< 1) <.Translator.translate>` calls.

    Examples:
        Basic usage.

        >>> btr = BitmaskTransformer(overrides={0b000: "NOT_SET", 0b1000: "OVERFLOW!"})

        Create a :class:`.Translator` using bitmask transforms for the `'bitmasks'` source.

        >>> from id_translation import Translator
        >>> data = {"id": [1, 4, 8], "name": ["name-of-1", "name-of-4", "0b1000"]}
        >>> tra = Translator({"bitmasks": data}, transformers={"bitmasks": btr})

        Translate some bitmasks!

        >>> tra.translate((0b000, 0b101, 8), names="bitmasks")
        ('NOT_SET', '1:name-of-1 & 4:name-of-4', 'OVERFLOW!')

        Note that ``0='NOT_SET'`` was translated even though it's not in the ``data``, and that ``8='0b1000'`` was
        replaced by ``'OVERFLOW!'``, as per the overrides specified for the transformer.

        Implication of setting ``force_real_translations=False``.

        >>> btr = BitmaskTransformer(force_real_translations=False)
        >>> tra = Translator({"bitmasks": data}, transformers={"bitmasks": btr})
        >>> tra.translate((5, 6), names="bitmasks", max_fails=0.0)
        ('1:name-of-1 & 4:name-of-4', '<Failed: id=2> & 4:name-of-4')

        The translation "succeeded", even though ``max_fails=0.0`` and ``6 = '<Failed: id=2> & 4:name-of-4'`` was only a
        partial success. This would've raised :class:`an error <.TooManyFailedTranslationsError>` if
        `force_real_translations` was not set. The transformer adds :attr:`~.MagicDict.real` mappings for all composite
        IDs, so the :class:`.Translator` won't detect any issues when using :meth:`.MagicDict.real_contains` to verify
        the results.
    """

    def __init__(
        self,
        joiner: str = " & ",
        *,
        overrides: _abc.Mapping[IdType, str] | None = None,
        force_decomposition: bool = False,
        force_real_translations: bool = True,
    ) -> None:
        self._joiner = joiner
        self._force = force_decomposition
        if isinstance(overrides, list):
            # TOML keys must be strings, so we use a record format.
            overrides = self._from_toml_records(overrides)
        self._overrides = overrides or {}
        self._force_real_translations = force_real_translations

    @classmethod
    def update_ids(cls, ids: set[IdType], /) -> None:
        """Add decomposed bitmask values."""
        new_ids = set()
        for decomposed in map(cls.decompose_bitmask, ids):
            new_ids.update(decomposed)
        ids.update(new_ids)

    def update_translations(self, translations: dict[IdType, str], /) -> None:
        """Join decomposed bitmask values using the `joiner` string."""
        ids_to_update = [idx for idx in translations if self.is_decomposable(idx)]
        if not self._force:
            ids_to_update = [idx for idx in ids_to_update if idx not in translations]

        translations.update(self._overrides)
        new_translations = {
            bitmask: self._create_composite_translation(self.decompose_bitmask(bitmask), translations=translations)
            for bitmask in ids_to_update
        }

        translations.update(new_translations)
        translations.update(self._overrides)

    def try_add_missing_key(self, key: IdType, /, *, translations: _abc.MutableMapping[IdType, str]) -> None:
        """Join decomposed bitmask values using the `joiner` string."""
        bits = self.decompose_bitmask(key)
        if not bits:
            return
        try:
            translations[key] = self._create_composite_translation(bits, translations=translations)
        except KeyError:
            return

    def _create_composite_translation(self, bits: list[IdType], *, translations: _abc.Mapping[IdType, str]) -> str:
        from id_translation.offline import MagicDict  # noqa: PLC0415

        if self._force_real_translations and isinstance(translations, MagicDict):
            translations = translations.real

        return self._joiner.join(translations[idx] for idx in bits)

    def __repr__(self) -> str:
        kwargs = dict(
            overrides=self._overrides,
            force_decomposition=self._force,
            force_real_translations=self._force_real_translations,
        )
        return f"{type(self).__name__}({self._joiner!r}, {_format_kwargs(kwargs)})"

    @classmethod
    def decompose_bitmask(cls, i: int, /) -> list[int]:
        """Decompose a bitmask into powers of two.

        If `i` is not :attr:`decomposable <is_decomposable>`, an empty list is returned.

        Args:
            i: Any integer.

        Returns:
            A decomposition of `i` into powers of two.
        """
        if not cls.is_decomposable(i):
            return []

        powers = []
        x = 1
        while x <= i:
            if x & i:
                powers.append(x)
            x <<= 1
        return powers

    @classmethod
    def is_decomposable(cls, i: int, /) -> bool:
        """Check if `i` is decomposable into bitmask values.

        An integer `i` is decomposable if and only if `i > 2`, and `i` is not a power of two.

        Args:
            i: Any integer.

        Returns:
            ``True`` if `i` is decomposable into powers of two.
        """
        return i > 2 and ((i & (-i)) != i)  # noqa: PLR2004

    @staticmethod
    def _from_toml_records(records: list[TomlOverrideRecord]) -> dict[IdType, str]:
        permitted = {"id", "override"}

        overrides = {}
        for i, record in enumerate(records, start=1):
            if permitted != set(record):
                msg = f"Record {i}/{len(records)} is malformed: Expected keys {permitted} but got {record}"
                raise ValueError(msg)

            key = record["id"]
            if key in overrides:
                raise ValueError(f"Duplicate ID in record {i}/{len(records)}: {record=}")
            overrides[key] = record["override"]
        return overrides



"""Functions and classes used by the :class:`.Mapper` for handling score matrices.

.. warning::

   This module is considered an implementation detail, and may change without notice.
"""

from warnings import warn as _warn

from ._score_helper import Record, Reject, ScoreHelper
from ._score_matrix import ScoreMatrix

__all__ = [
    "Record",
    "Reject",
    "ScoreHelper",
    "ScoreMatrix",
]

_warn(
    "This module is considered an implementation detail, and may change without notice.",
    UserWarning,
    stacklevel=2,
)
del _warn



"""Functions which return a likeness score.

See Also:
    The :class:`~.HeuristicScore` class.
"""

from collections.abc import Iterable as _Iterable

from . import exceptions
from .types import CandidateType, ContextType, ValueType


def modified_hamming(
    name: str,
    candidates: _Iterable[str],
    context: ContextType | None,  # noqa: ARG001
    *,
    add_length_ratio_term: bool = True,
    positional_penalty: float = 0.001,
) -> _Iterable[float]:
    """Compute hamming distance modified by length ratio, from the back. Score range is ``[0, 1]``.

    Args:
        name: A name that should be mapped one of the sources in `candidates`.
        candidates: Candidate sources.
        context: Should be ``None``. Always ignored, exists for compatibility.
        add_length_ratio_term: If ``True``, score is divided by ``abs(len(name) - len(candidate))``.
        positional_penalty: A penalty applied to prefer earlier `candidates`, according to the formulare
            ``penalty = index(candidate) * positional_penalty)``.

    Examples:
        >>> from id_translation.mapping.score_functions import modified_hamming
        >>> list(modified_hamming("aa", ["aa", "a", "ab", "aa"], context=None))
        [1.0, 0.499, 0.498, 0.997]
        >>> list(
        ...     modified_hamming(
        ...         "aa", ["aa", "a", "ab", "aa"], context=None, positional_penalty=0
        ...     )
        ... )
        [1.0, 0.5, 0.5, 1.0]
        >>> list(modified_hamming("face", ["face", "FAce", "race", "place"], context=None))
        [1.0, 0.499, 0.748, 0.372]
    """

    def _apply(candidate: str) -> float:
        sz = min(len(candidate), len(name))
        same = sum([name[i] == candidate[i] for i in range(-sz, 0)])

        ratio = (1 / (1 + abs(len(candidate) - len(name)))) if add_length_ratio_term else 1
        normalized_hamming = same / sz

        return ratio * normalized_hamming

    yield from (s - i * positional_penalty for i, s in enumerate(map(_apply, candidates)))


def equality(
    value: ValueType,
    candidates: _Iterable[CandidateType],
    context: ContextType | None,  # noqa: ARG001
) -> _Iterable[float]:
    """Return 1.0 if ``k == c_i``, 0.0 otherwise.

    Examples:
        >>> from id_translation.mapping.score_functions import equality
        >>> list(equality("a", "aAb", context=None))
        [1.0, 0.0, 0.0]
    """
    yield from map(float, (value == c for c in candidates))


def disabled(
    value: ValueType,
    candidates: _Iterable[CandidateType],
    context: ContextType | None,
    strict: bool = True,
) -> _Iterable[float]:
    """Special value to indicate that scoring logic has been disabled.

    This is a workaround to allow users to indicate that the scoring logic is disabled, and that overrides should be
    used instead. The ``disabled``-function has no special meaning to the mapper, and will be called as any other
    scoring function.

    Returns:
        If `strict` is ``False``, negative infinity for all `candidates`, serving as a catch-all removal filter.

    Raises:
        ScoringDisabledError: If `strict` is ``True``.

    See Also:
        The :ref:`override-only-mapping` documentation.
    """
    if strict:
        raise exceptions.ScoringDisabledError(value, candidates, context)

    return [float("-inf")] * sum(1 for _ in candidates)
